# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Shared algorithms which could be reused or optimised."""

import math
from typing import Callable, List, Optional, Sequence, TypeVar, Union
import numpy  # type: ignore
from ..types import (Line2D, Line3D, LineString2D, LineString3D, Point2D,
                     Point3D, Polygon2D, Rectangle2D, Triangle3D)

T = TypeVar('T')


def counting_sort(input_list: Sequence[T],
                  key: Callable[[T], int] = int) -> List[T]:  # type: ignore
    """Sort a sequence by an integer key.

    This creates a new list from the input sequence. The callable will
    be called several times for each value in the sequence, make sure
    it is as light-weight as possible.

    Parameters
    ----------
    input_list : Sequence[T]
        The input sequence to sort
    key : Callable[[T], int], optional
        The key function to use for converting the list of elements
        into integer keys, by default int

    Returns
    -------
    List[T]
        A copy of the input sequence, sorted by the given key
    """
    # NOTE: Regarding the ignore at the function definition: Mypy complains
    # about "int" not being a callable as it interprets it as a type. It is a
    # callable though, as it has a constructor. This error is being silenced.

    # Find the minimum and maximum key values. The minimum value is used to
    # shift the list key values to allow for negative keys in the input
    min_key = max_key = key(input_list[0])
    for element in input_list:
        value = key(element)
        if value > max_key:
            max_key = value
        elif value < min_key:
            min_key = value
    slots = max_key - min_key + 1

    # Create an array large enough to accomodate the entire value range
    indices = list(0 for _ in range(slots))
    # Populate the array with the number of occurrences of each key value
    for element in input_list:
        indices[key(element) - min_key] += 1
    # Shift each key starting occurrence up by the number of preceding
    # occurrences
    for index, _ in enumerate(indices):
        try:
            indices[index+1] += indices[index]
        except IndexError:
            continue

    # Generate the output list, moving each element in the input to the
    # position in the output list designated by the indices list
    output_list: List[Optional[T]] = list(None for _ in input_list)
    for element in input_list:
        value_shifted = key(element) - min_key
        # Add the element
        output_list[indices[value_shifted] - 1] = element
        # Decrement the index
        indices[value_shifted] -= 1

    # Discard any None values
    return [t for t in output_list if t is not None]


def dist_2d(pt_a: Point2D, pt_b: Point2D) -> float:
    """Return the Euclidean 2D distance between two points.

    Parameters
    ----------
    pt_a : Tuple[float, float]
        The first input point
    pt_b : Tuple[float, float]
        The second input point

    Returns
    -------
    float
        The straight-line distance between the two points
    """
    dist_x = pt_a[0] - pt_b[0]
    dist_y = pt_a[1] - pt_b[1]
    return math.sqrt(dist_x**2 + dist_y**2)


def dist_3d(pt_a: Point3D, pt_b: Point3D) -> float:
    """Return the Euclidean 3D distance between two points.

    Like dist_2d, but this version also considers differences in
    elevation.

    Parameters
    ----------
    pt_a : Tuple[float, float, float]
        The first input point
    pt_b : Tuple[float, float, float]
        The second input point

    Returns
    -------
    float
        The straight-line distance between the two points
    """
    dist_x = pt_a[0] - pt_b[0]
    dist_y = pt_a[1] - pt_b[1]
    dist_z = pt_a[2] - pt_b[2]
    return math.sqrt(dist_x**2 + dist_y**2 + dist_z**2)


def get_intersections(line: Line2D, polygon: Polygon2D) -> List[Point2D]:
    """Find all intersections between the given line and polygon.

    This returns a list of all points where the given line intersects
    with the edges of the polygon. Coincident lines are not considered
    "real" intersections and will not be part of the return list.

    Parameters
    ----------
    line : Tuple[Tuple[float, float], Tuple[float, float]]
        The line to check for intersections
    polygon : Tuple[Tuple[float, float], ...]
        A polygon whos edges will be checked for intersections. This
        must have at least three vertices.

    Returns
    -------
    List[Tuple[float, float]]
        The list of intersections

    Raises
    ------
    ValueError: Raised if the given polygon has fewer than three
        vertices.
    """
    if len(polygon) < 3:
        raise ValueError('The polygon must have at least 3 vertices')
    intersections: List[Point2D] = []
    for index, vtx_1 in enumerate(polygon):
        try:
            vtx_2 = polygon[index+1]
        except IndexError:
            vtx_2 = polygon[0]
        edge = vtx_1, vtx_2
        # Check for an intersection with the given line
        if not line_segments_intersection(line, edge, False):
            continue
        # If an intersection was expected, calculated its coordinates
        intersect = line_intersection(line, edge)
        intersections.append(intersect)
    return intersections


def interpolate_line(point: Point2D, line: Line3D) -> float:
    """Interpolate the elevation of a line.

    The given point will be snapped onto the line, it does not have to
    be exactly on the line.

    Parameters
    ----------
    point : Tuple[float, float]
        The point at which to calculate the elevation of the 3D line
    line : Tuple[Tuple[float, float, float],
                 Tuple[float, float, float]]
        The 3D line to use for interpolation

    Returns
    -------
    float
        The elevation of the line at the point closest to the input
        point
    """
    # Project the point onto the line to get a relative distance along the line
    line_vec = numpy.array((line[1][0]-line[0][0], line[1][1]-line[0][1]))
    point_vec = numpy.array((point[0]-line[0][0], point[1]-line[0][1]))
    len_sq = dist_2d(line[1][:2], line[0][:2]) ** 2

    # Project the point
    rel_dist = line_vec @ point_vec / len_sq
    # Interpolate the new height
    return float(line[0][2] + rel_dist * (line[1][2]-line[0][2]))


def interpolate_triangle(point: Point2D, triangle: Triangle3D) -> float:
    """Interpolate the elevation of a triangle.

    This returns the elevation of the plane defined by the given
    triangle at the location of the point.

    This uses Barycentric coordinates; the point does not need to lie
    in or even near the triangle itself.

    Parameters
    ----------
    point : Tuple[float, float]
        The point to interpolate
    triangle : Tuple[Tuple[float, float, float],
                     Tuple[float, float, float],
                     Tuple[float, float, float]]
        The triangle to use for interpolation

    Returns
    -------
    float
        The interpolated elevation of the triangle at the point
    """
    # NOTE: Barycentric coordinates do support negative weights, which could be
    # used to check for point containment within triangles. Might be worth
    # looking into if we're chasing performance.

    vtx_1, vtx_2, vtx_3 = triangle

    # Calculate Barycentric weights for each vertex
    denom = ((vtx_2[1] - vtx_3[1]) * (vtx_1[0] - vtx_3[0])
             + ((vtx_3[0] - vtx_2[0]) * (vtx_1[1] - vtx_3[1])))

    weight_1 = ((vtx_2[1] - vtx_3[1]) * (point[0] - vtx_3[0])
                + ((vtx_3[0] - vtx_2[0]) * (point[1] - vtx_3[1]))) / denom

    weight_2 = ((vtx_3[1] - vtx_1[1]) * (point[0] - vtx_3[0])
                + ((vtx_1[0] - vtx_3[0]) * (point[1] - vtx_3[1]))) / denom

    weight_3 = 1 - weight_1 - weight_2

    # The interpolated value is now equal to the sum of all weights and their
    # corresponding input values
    return weight_1*vtx_1[2] + weight_2*vtx_2[2] + weight_3*vtx_3[2]


def half_plane_distance(line: Line2D, point: Point2D,
                        normalise: bool = True) -> float:
    """Return the signed distance from the point to the line.

    This returns the signed half plane distance of the point from the
    line.
    The line bisects 2D plane into two half-planes. If the point lies
    on this line, the distance returned is 0.0.
    If the point lies to the left of an observer sitting at the first
    point of the line and looking towards the second point, the
    distance returned will be negative. If the point lies to the right
    of this observer, the distance will be positive.

    By default, the returned distance is normalised; i.e. it will equal
    the actual Euclidean distance between the point and the line's
    closest approach.
    If the sign is more important than the actual value, you may set
    the normalise argument to False to save yourself a math.sqrt()
    call.

    Parameters
    ----------
    line : Tuple[Tuple[float, float], Tuple[float, float]]
        The line splitting the 2D plane
    point : Tuple[float, float]
        The point to calculate the half plane distance for
    normalise : bool, optional
        Whether to normalise the returned distance, by default True

    Returns
    -------
    float
        The signed distance from the point to the line
    """
    # NOTE: I cannot find any convention on which side of the 2D plane should
    # be represented by which sign.
    # I opted for this arrangement so that running this check for a polygon
    # with CCW vertex order returns sensible (i.e. positive) distances for
    # external points and negative distances for internal points.
    # The distance to an object the point is contained in seemed more deserving
    # of the negative sign.

    # Build a vector from the given input line
    vec = numpy.array((line[1][0]-line[0][0], line[1][1]-line[0][1]))
    # Get the right-hand normal vector of the line
    norm_vec = numpy.array((vec[1], -vec[0]))
    # Get the unit vector (this preserves Euclidean distance in the output)
    if normalise:
        norm_vec = norm_vec / dist_2d(*line)
    # Get a vector representing the input point
    check_vec = numpy.array((point[0]-line[0][0], point[1]-line[0][1]))

    # Project the check-vector onto the edge normal vector. If the projection
    # is negative, the point lies on the "inside" of the edge.
    return float(norm_vec @ check_vec)


def distance_to_polygon(point: Point2D, polygon: Polygon2D) -> float:
    """Calculate the distance from a convex polygon to a given point.

    The return value will be negative if the point lies within the
    polygon.

    Due to the multiple reprojections used, this algorithm is highly
    susceptible to floating point errors. This means that points lying
    on the edge of the polygon will hardly ever return 0.0, but rather
    something like +/- 1.0e-16.
    It is recommended to run a separate point-on-line test for these
    values if you are interested in the precise value.

    Note that this algorithm does not check whether the given polygon
    is actually convex. It also does not check for self-intersections
    or verifies whether the vertices are ordered in counter-clockwise
    order.

    Parameters
    ----------
    point : Tuple[float, float]
        The point to calculate the distance for
    polygon : Tuple[Tuple[float, float], ...]
        Any number of vertices forming a convex polygon; at least three
        vertices are required

    Returns
    -------
    float
        The shortest distance from the point to any polygon edge; will
        be nagative if the point lies within the polygon

    Raises
    ------
    ValueError
        Raised if the given polygon has fewer than three vertices
    """
    if len(polygon) < 3:
        raise ValueError('A valid polygon must have at least three '
                         f'vertices (got {len(polygon)})')

    # NOTE: This is identical in principle to the point_in_polygon_convex
    # algorithm, but the loop doesn't end once the point is deemed to lie
    # outside the polygon.
    # Instead, every point is projected back onto the edge it was checked
    # against and constrained onto its length. This also implicitly checks the
    # vertices.

    # Used to keep track of the minimum distance encountered
    min_dist = None

    # Iterate over the vertices of the polygon
    for index, vtx_1 in enumerate(polygon):
        # Find a second, neighbouring vertex
        try:
            vtx_2 = polygon[index+1]
        except IndexError:
            # Connect the last vertex to the first, closing the ring
            vtx_2 = polygon[0]

        # Build a vector from the given edge
        edge_vec = numpy.array((vtx_2[0]-vtx_1[0], vtx_2[1]-vtx_1[1]))
        # Build a vector from the first vertex to the point in question
        check_vec = numpy.array((point[0]-vtx_1[0], point[1]-vtx_1[1]))

        # Calculate the squared length of the edge (this avoids a sqrt later)
        len_sq = dist_2d(vtx_1, vtx_2) ** 2

        # Project the check-vector onto the edge vector and calculate the
        # relative distance (i.e. distance along the edge)
        rel_dist = edge_vec @ check_vec / len_sq

        # Clamp the relative distance. This ensures that a point too far along
        # the vector end sup being closest to one of the two end points of the
        # edge
        if rel_dist < 0.0:
            rel_dist = 0.0
        elif rel_dist > 1.0:
            rel_dist = 1.0

        # Interpolate the point along the edge that is closest to the input
        # point
        pos_x = vtx_1[0] + edge_vec[0] * rel_dist
        pos_y = vtx_1[1] + edge_vec[1] * rel_dist

        # Finally, calculate the distance between the input point and the
        # clamped, projected point
        dist = dist_2d(point, (pos_x, pos_y))

        if min_dist is None or dist < min_dist:
            min_dist = dist

    # At this point of the loop, min_dist has been set to a float
    assert min_dist is not None

    # If the point lies inside the polygon, return a negative number
    if point_in_polygon_convex(point, polygon):
        min_dist = - float(min_dist)
    return min_dist


def line_intersection(line_1: Line2D, line_2: Line2D) -> Point2D:
    """Return the intersection point of two intersecting lines.

    Note that this treats both lines as infinite rays, the intersection
    may not lie on either input line.

    Parameters
    ----------
    line_1 : Tuple[Tuple[float, float], Tuple[float, float]]
        The first input line
    line_2 : Tuple[Tuple[float, float], Tuple[float, float]]
        The second input line

    Returns
    -------
    Tuple[float, float]
        The point of intersection

    Raises
    ------
    ValueError
        Raised if the two lines are parellel or collinear
    """
    pt_a, pt_b = line_1
    pt_c, pt_d = line_2

    # Succeed early if the lines share an endpoint
    if pt_a in line_2:
        return pt_a
    if pt_b in line_2:
        return pt_b

    # Convert the lines to the implicit form
    a_1, b_1 = pt_b[1]-pt_a[1], -(pt_b[0]-pt_a[0])
    a_2, b_2 = pt_d[1]-pt_c[1], -(pt_d[0]-pt_c[0])
    d_1 = a_1*pt_a[0] + b_1*pt_a[1]
    d_2 = a_2*pt_c[0] + b_2*pt_c[1]

    # Calculate intersection
    denom = a_1*b_2 - a_2*b_1

    if denom == 0:
        raise ValueError('No intersection found, lines are parallel')

    pos_x = (b_2*d_1 - b_1*d_2) / denom
    pos_y = (a_1*d_2 - a_2*d_1) / denom

    return pos_x, pos_y


def line_segments_intersection(line_1: Line2D, line_2: Line2D,
                               allow_collinear: bool = True) -> bool:
    """Return whether the two given lines intersect.

    This returns a Boolean value. Use line_intersection to retrieve the
    actual point of intersection if this succeeds.

    Parameters
    ----------
    line_1 : Tuple[Tuple[float, float], Tuple[float, float]]
        The first line segment
    line_2 : Tuple[Tuple[float, float], Tuple[float, float]]
        The second line segment
    allow_collinear : bool, optional
        Whether collinear lines count as an intersection, by default
        True

    Returns
    -------
    bool
        Whether the two line segments intersect
    """
    pt_a, pt_b = line_1
    pt_c, pt_d = line_2

    # Fail early if the two lines' bounding rectangles don't overlap
    min_x_a, max_x_a = ((pt_a[0], pt_b[0]) if pt_a[0] <= pt_b[0]
                        else (pt_b[0], pt_a[0]))
    min_y_a, max_y_a = ((pt_a[1], pt_b[1]) if pt_a[1] <= pt_b[1]
                        else (pt_b[1], pt_a[1]))
    min_x_b, max_x_b = ((pt_c[0], pt_d[0]) if pt_c[0] <= pt_d[0]
                        else (pt_d[0], pt_c[0]))
    min_y_b, max_y_b = ((pt_c[1], pt_d[1]) if pt_c[1] <= pt_d[1]
                        else (pt_d[1], pt_c[1]))
    rect_a = min_x_a, max_x_a, min_y_a, max_y_a
    rect_b = min_x_b, max_x_b, min_y_b, max_y_b
    if not rectangle_intersection(rect_a, rect_b):
        return False

    # Succeed early if the two lines share a point
    if (pt_a in line_2 or pt_b in line_2) and allow_collinear:
        return True

    # Calculate the vectors for each line
    vec_1 = numpy.array((pt_b[0]-pt_a[0], pt_b[1]-pt_a[1]))
    vec_2 = numpy.array((pt_d[0]-pt_c[0], pt_d[1]-pt_c[1]))
    vec_3 = numpy.array((pt_c[0]-pt_a[0], pt_c[1]-pt_a[1]))

    vec_3cross1 = numpy.cross(vec_3, vec_1)
    vec_3cross2 = numpy.cross(vec_3, vec_2)
    vec_1cross2 = numpy.cross(vec_1, vec_2)

    # Check for collinearity
    if vec_3cross1 == 0:

        if allow_collinear:
            # Return whether there is any overlap betweeen the two lines
            return ((pt_c[0]-pt_a[0] < 0) != (pt_c[0]-pt_b[0])
                    or (pt_c[1]-pt_a[1] < 0) != (pt_c[1]-pt_b[1]))

        return False

    # Check whether the two lines are parallel (but not collinear)
    if vec_1cross2 == 0:
        return False

    # Express the intersection using each line's parametric representation
    scalar = 1 / vec_1cross2
    para_1 = vec_3cross1 * scalar
    para_2 = vec_3cross2 * scalar

    # The intersection lies on the two lines only if the parameter falls within
    # the [0, 1] range for both lines
    return bool(0 <= para_1 <= 1 and 0 <= para_2 <= 1)


def point_in_polygon_concave(point: Point2D, polygon: Polygon2D) -> bool:
    """Return whether the given point lies within the polygon.

    Note that this algorithm does not check for self-intersection.

    Parameters
    ----------
    point : Tuple[float, float]
        The point to check
    polygon : Tuple[Tuple[float, float], ...]
        The polygon to check against, this must have at least three
        vertices

    Returns
    -------
    bool
        Whether the point lies within the polygon

    Raises
    ------
    ValueError
        Raised if the polygon has fewer than three vertices
    """
    # NOTE: The following uses a ray-casting algorithm, i.e. works by casting a
    # horizontal ray from the point in question and counting the number of
    # intersections with the polygon boundary.
    if len(polygon) < 3:
        raise ValueError('A valid polygon must have at least three '
                         f'vertices (got {len(polygon)})')

    is_inside = False

    for index, vtx_1 in enumerate(polygon):
        try:
            vtx_2 = polygon[index+1]
        except IndexError:
            vtx_2 = polygon[0]

        # NOTE: The rest of the algorithm cannot handle points sitting on an
        # edge and will return arbitrary values
        if point_on_line((vtx_1, vtx_2), point):
            return True

        # If the edge crosses the ray (either direction)
        if (vtx_1[1] < point[1] <= vtx_2[1]
                or vtx_2[1] < point[1] <= vtx_1[1]):
            # If the point of intersection lies to the left of the point, flip
            # the output flag
            if (vtx_2[0] + (point[1]-vtx_2[1]) / (vtx_1[1]-vtx_2[1])
                    * (vtx_1[0]-vtx_2[0]) < point[0]):
                is_inside = not is_inside
    return is_inside


def point_in_polygon_convex(point: Point2D, polygon: Polygon2D) -> bool:
    """Return whether the given point lies within the polygon.

    Note that this algorithm does not check whether the polygon is
    actually convex, whether it has any self-intersection or whether
    its vertices are ordered in counter-clockwise order.

    Parameters
    ----------
    point : Tuple[float, float]
        The point to check
    polygon : Tuple[Tuple[float, float], ...]
        The polygon to check against, this must have at least three
        vertices

    Returns
    -------
    bool
        Whether the point lies within the given polygon

    Raises
    ------
    ValueError
        Raised if the polygon has fewer than three vertices
    """
    if len(polygon) < 3:
        raise ValueError('A valid polygon must have at least three '
                         f'vertices (got {len(polygon)})')

    # Iterate over the vertices of the polygon
    for index, vtx_1 in enumerate(polygon):
        # Find a second, neighbouring vertex
        try:
            vtx_2 = polygon[index+1]
        except IndexError:
            # Connect the last vertex to the first, closing the ring
            vtx_2 = polygon[0]

        # Get the signed distance of the current edge and point. If it is
        # positive, the point lies on the outside half plane for this edge and
        # can therefore not be part of the polygon any more.
        if half_plane_distance((vtx_1, vtx_2), point, normalise=False) > 0.0:
            return False

    # If the point lies on the "inside" of all edges of a convex polygon, it
    # lies within the polygon
    return True


def point_on_line(line: Line2D, point: Point2D,
                  precision: float = 0.0) -> bool:
    """Return whether the given point lies on the line.

    Due to floating point inaccuracies, you may want to set the
    precision argument to a sensible value when working with diagonal
    lines.

    Parameters
    ----------
    line : Tuple[Tuple[float, float], Tuple[float, float]]
        The input line to check against
    point : Tuple[float, float]
        The point to check
    precision : float, optional
        The maximum distance the point may be from the line to still be
        considered on the line, by default 0.0

    Returns
    -------
    bool
        Whether the point lies on the line, within tolerance
    """
    # Fail early if the point lies outside the line's bounding rectangle
    if not ((line[0][0] <= point[0] <= line[1][0]
             or line[1][0] <= point[0] <= line[0][0])
            and (line[0][1] <= point[1] <= line[1][1]
                 or line[1][1] <= point[1] <= line[0][1])):
        return False
    line_vec = numpy.array((line[1][0]-line[0][0], line[1][1]-line[0][1]))
    point_vec = numpy.array((point[0]-line[0][0], point[1]-line[0][1]))
    # The span of two parallel vectors must be one-dimensional
    if (abs(numpy.cross(line_vec, point_vec))
            / (line_vec @ point_vec) > precision):
        return False
    # Project the point onto the line
    proj = line_vec @ point_vec
    # If the projection is negative, the point lies before the line
    return bool(-precision <= proj <= dist_2d(*line) ** 2 + precision)


def point_within_range(point_a: Point2D, point_b: Point2D,
                       max_dist: float) -> bool:
    """Return whether two points lie within a given distance.

    This uses a two step search, first evaluating the X and Y
    coordinates individually before performing the actual Euclidean
    distance calculation, which allows it to fail early for faraway
    points.

    Parameters
    ----------
    point_a : Tuple[float, float]
        The first point
    point_b : Tuple[float, float]
        The second point
    max_dist : float
        The maximum distance between the two points

    Returns
    -------
    bool
        Whether the distance between the two points is within range
    """
    # Step 1: coordinate comparison (square search area)
    if abs(point_a[0]-point_b[0]) > max_dist:
        return False
    if abs(point_b[1]-point_b[1]) > max_dist:
        return False
    # Step 2: Euclidean distance (circular search area)
    return dist_2d(point_a, point_b) <= max_dist


def rectangle_intersection(rect_a: Rectangle2D, rect_b: Rectangle2D) -> bool:
    """Return whether the two rectangles overlap each other."""
    # NOTE: Rectangle2D : Tuple[min_x, max_x, min_y, max_y]
    return (rect_b[0] <= rect_a[1]
            and rect_b[1] >= rect_a[0]
            and rect_b[2] <= rect_a[3]
            and rect_b[3] >= rect_a[2])


def rotate_2d(point: Point2D, angle: float, anchor: Point2D) -> Point2D:
    """Rotate the given point around an anchor point."""
    radians = angle * (math.pi/180)
    dist_x = point[0] - anchor[0]
    dist_y = point[1] - anchor[1]
    new_x = math.cos(radians) * dist_x - math.sin(radians) * dist_y + anchor[0]
    new_y = math.sin(radians) * dist_x + math.cos(radians) * dist_y + anchor[1]
    return new_x, new_y


def split_line(line_string: Union[LineString2D, LineString3D],
               num_segments: int) -> Union[LineString2D, LineString3D]:
    """Split the given line string into a number of segments.

    The start and end point remain unchanged, but additional
    equidistant points are inserted to achieve the desired number of
    subsegments.

    Note that for line strings with more than two points, this will
    split each of the lines. A line string with three points with a
    num_segments of 2 will result in an output line string with five
    points.

    If the input line is 3D, the inserted points will be interpolated
    to match the line's elevation.

    Parameters
    ----------
    line_string : Tuple[Union[Tuple[float, float],
                              Tuple[float, float, float]], ...]
        A line string consisting of 2D or 3D points
    num_segments : int
        The number of subsegments to split each line into, must be at
        least 1

    Returns
    -------
    Tuple[Union[Tuple[float, float], Tuple[float, float, float]], ...]
        The input line string, with additional vertices added

    Raises
    ------
    ValueError
        Raised if num_segments is less than 1
    """
    if num_segments < 1:
        raise ValueError('num_segments must greater or equal to one')
    assert len(line_string) >= 2, 'Line strings require at least two points'

    # Process each line individually
    out_string: List[Union[Point2D, Point3D]] = []
    for index, start_point in enumerate(line_string[:-1]):
        end_point = line_string[index+1]

        delta_x = end_point[0] - start_point[0]
        delta_y = end_point[1] - start_point[1]

        # Add the start point and any other inner points
        for cut_idx in range(num_segments):
            pos_x = start_point[0] + delta_x * (cut_idx/num_segments)
            pos_y = start_point[1] + delta_y * (cut_idx/num_segments)
            new_pt: Union[Point2D, Point3D] = (pos_x, pos_y)

            # Add elevation if needed
            if len(start_point) >= 3:
                pt_2d: Point2D = (new_pt[0], new_pt[1])
                # NOTE: If the code gets here, start_point and end_point are
                # going to be 3D points.
                pos_z = interpolate_line(
                    pt_2d, (start_point, end_point))  # type: ignore
                new_pt = *pt_2d, pos_z

            out_string.append(new_pt)

        # If we're at the end of the line string, add the end point as well
        if end_point == line_string[-1]:
            out_string.append(end_point)

    return tuple(out_string)  # type: ignore



#!/usr/bin/env python

# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Command line script for the BASEchange channel generator.

This script exposes the core functionality of the BASEchange submodule
and allows the creation of 1D channel geometries from the given input
parameters.

Run this script with the "--help" for additional information and
detailed argument descriptions.
"""

import os
import subprocess
import sys

try:
    # pylint: disable=import-error, unused-import
    import basemesh.basechange
except ModuleNotFoundError as err:
    basemesh_dir, _ = os.path.split(os.path.dirname(os.path.abspath(__file__)))
    sys.path.append(basemesh_dir)
    try:
        import basemesh.basechange
    except ModuleNotFoundError:
        raise err from err
else:
    basemesh_dir = os.path.dirname(__file__)


if __name__ == '__main__':
    # Update PYTHONPATH
    path = os.environ.get('PYTHONPATH', '')
    if path:
        path += ':'
    path += os.path.abspath(basemesh_dir)
    os.environ['PYTHONPATH'] = path
    # Get script path
    subprocess.call(
        (sys.executable, '-m', 'basemesh', '--1D', *sys.argv[1:]))



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Worker objects for long-running operations."""

from collections import deque
from contextlib import redirect_stdout
from typing import Any, Callable, Deque, Iterable, IO, Mapping, Optional
from PyQt5.QtCore import QObject, pyqtSignal
from ...feedback import Feedback


class Worker(QObject):
    """Generic worker task wrapper."""

    done = pyqtSignal(object)  # The return value of the worker
    error = pyqtSignal(object)  # The error raised
    print = pyqtSignal(object)  # The deque containing the printed data
    status = pyqtSignal(object)  # The feedback status tuple

    def __init__(self, target: Callable[..., Any],
                 args: Optional[Iterable[Any]] = None,
                 kwargs: Optional[Mapping[str, Any]] = None) -> None:
        """Initialise the worker for a given callable.

        Parameters
        ----------
        target : Callable[..., Any]
            The callable to run in the worker thread
        args : Iterable[Any], optional
            Any positional arguments to pass to the target callable, by
            default None
        kwargs : Mapping[str, Any], optional
            Any keyword arguments to pass to the target callable, by
            default None
        """
        super().__init__()
        self.args = args if args is not None else ()
        self.feedback = Feedback(self._emit_status)
        self.kwargs = kwargs if kwargs is not None else {}
        self.target = target
        self._queue: Deque[str] = deque()

    def _emit_status(self, progress: Optional[float], msg: Optional[str],
                     busy: bool) -> None:
        """Emit the status signal for this worker.

        Parameters
        ----------
        progress : Optional[float]
            The new progress value, or None to keep unchanged
        msg : Optional[str]
            The new status message, or None to keep unchanged
        busy : bool
            Whether the worker is currently busy
        """
        self.status.emit((progress, msg, busy))

    def cancel(self) -> None:
        """Flag to inform the worker to stop operation."""
        self.feedback.cancel()

    def run(self) -> None:
        """Run the worker's target callable.

        This is the task that will be spun off into another thread.
        """
        # NOTE: This stream technically does not support the IO ABC, but adding
        # all of those methods with NotImplementedErrors just for type
        # strictness seems a bit pointless.
        stream: IO[str] = _WriteOnlyStream(
            self.print, self._queue)  # type: ignore
        with redirect_stdout(stream):
            try:
                value = self.target(self.feedback, *self.args, **self.kwargs)
            except BaseException as err:  # pylint: disable=broad-except
                self.error.emit(err)
                value = None
        self.done.emit(value)


class _WriteOnlyStream:
    """A write-only text stream used to redirect prints to the caller.

    This does not follow the IO[str] type due to missing
    functionality and is only intended for internal use with the
    contextlib.redirect_stdout() context manager.
    """

    def __init__(self, signal: pyqtSignal, output: Deque[str]) -> None:
        """Instantiate a new write-only, cross-thread stream.

        Parameters
        ----------
        signal : pyqtSignal
            The signal to emit when written to
        output : Deque[str]
            An external deque to store writes in
        """
        self._output = output
        self._signal = signal

    def close(self) -> None:
        """Close the stream."""
        self._output.clear()

    def write(self, msg: str) -> int:
        """Write the given string to the stream.

        Parameters
        ----------
        msg : str
            The message to write

        Returns
        -------
        int
            The number of characters written
        """
        self._output.append(msg)
        self._signal.emit(self._output)
        return len(msg)



#!/usr/bin/env python

# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Command line script for the BASEchange channel generator.

This script exposes the core functionality of the BASEchange submodule
and allows the creation of 1D channel geometries from the given input
parameters.

Run this script with the "--help" for additional information and
detailed argument descriptions.
"""

import argparse
from typing import List

from basemesh.basechange import (factories, BaseChainWriter,
                                 __version__ as version)
from basemesh.triangle import RegionMarker

SKETCH = r"""
  
  ┌─ DX ─┐      ┌ DXT ┐    ┌─ DXW ─┐
  |      |      |     |                     
  |      |      |     |    *****************
  |      |      |      ****·   |   ·       ·****     
                   ****    ·   |   ·       ·    ****
  *****************   ·    ·   |   ·       ·    ·   ****************
  ·  |   ·      ·     ·    ·   |   ·       ·    ·    ·  |   ·      ·
  ·  WC  ·      ·     ·    ·   WW  ·       ·    ·    ·  WC  ·      ·
  ·  |   ·      ·     ·    ·   |   ·       ·    ·    ·  |   ·      ·
  *****************   ·    ·   |   ·       ·    ·   ****************
                   ****    ·   |   ·       ·    *\**
   \           /       ****·   |   ·       ·****  \
    \___ N ___/            *****************       \
                                                 NT: Number of interpolated
                            \             /      transition cross sections
                             \___ NW ____/      
 
"""


if __name__ == '__main__':
    print('\n'
          '---------------------------------------\n'
          f'BASEchange - Version {version}\n'
          '---------------------------------------\n')

    # Set up argument parser
    parser = argparse.ArgumentParser(
        'BASEchange', description=SKETCH,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument(
        '--version', action='version')

    # Core options
    group = parser.add_argument_group(
        'channel definition', description='Core options for defining a '
        '1D channel geometry')

    group.add_argument(
        '-DX', metavar='dx', dest='dx', default=40.0, type=float,
        help='inter-cross-sectional distance for the channel '
             '(default %(default)s)')
    group.add_argument(
        '-N', metavar='num_cs', dest='num_cs', default=10, type=int,
        help='number of cross sections for the channel. When setting the '
             '--widening flag, this value is used for the first and last '
             'section of the channel (default %(default)s)')
    group.add_argument(
        '-S', metavar='slope_channel', dest='slope_channel', default='0.0015',
        type=float, help='channel slope [m] (default %(default)s)')
    group.add_argument(
        '-WC', metavar='bed_width', dest='bed_width', default=50.0, type=float,
        help='channel bed width [m] (default: %(default)s)')
    group.add_argument(
        '-B', metavar='slope_banks', dest='slope_banks', default='1.0',
        type=float, help='slope of channel banks [h:v] (default: %(default)s)')
    group.add_argument(
        '-H', metavar='height', dest='height', default=5.0, type=float,
        help='height of the floodplain relative to the channel bed [m] '
             '(default: %(default)s)')

    # Frictions
    group.add_argument(
        '-KS', metavar='ks_bed', dest='ks_bed', default=35.0, type=float,
        help='channel bed Strickler friction value [m^(1/3)/s] '
             '(default %(default)s)')
    group.add_argument(
        '-KSB', metavar='ks_banks', dest='ks_banks', default=20.0, type=float,
        help='channel banks Strickler friction value [m^(1/3)/s] '
             '(default %(default)s)')

    # Offsets
    group.add_argument(
        '-X', metavar='offset_x', dest='offset_x', default=0.0, type=float,
        help='channel offset along flow axis [m] (default: %(default)s)')
    group.add_argument(
        '-Z', metavar='offset_z', dest='offset_z', default=0.0, type=float,
        help='reference elevation for the channel bed [m] '
             '(default: %(default)s)')
    group.add_argument(
        '-MPS', metavar='mp_shift', dest='mp_shift', default=0.5, type=float,
        help='relative position of the channel bank midpoint '
             '(default: %(default)s)')

    # Widening options
    group = parser.add_argument_group(
        'widening options', description='Additional options for channels '
        'featuring a widening')
    group.add_argument(
        '--W', '--widening', dest='widening', action='store_true',
        help='include widening; other arguments in this group are only used '
             'if this flag was set (default %(default)s)')
    group.add_argument(
        '-WW', metavar='w_bed_width', dest='w_bed_width', default=100.0,
        type=float, help='bed width of the widened section [m] '
        '(default: %(default)s)')
    group.add_argument(
        '-KSW', metavar='w_ks_bed', dest='w_ks_bed', default=30.0, type=float,
        help='widening and transitionings Strickler value [m^(1/3)/s] '
             '(default %(default)s)')
    group.add_argument(
        '-DXW', metavar='w_dx', dest='w_dx', default=30.0, type=float,
        help='inter-cross-sectional distance for the widening '
             '(default %(default)s)')
    group.add_argument(
        '-NW', metavar='w_num_cs', dest='w_num_cs', default=5, type=int,
        help='number of cross sections for the widening '
             '(default %(default)s)')
    group.add_argument(
        '-DXT', metavar='t_dx', dest='t_dx', default=20.0, type=float,
        help='inter-cross-sectional distance for the transitions '
             '(default %(default)s)')
    group.add_argument(
        '-NT', metavar='t_num_cs', dest='t_num_cs', default=3, type=int,
        help='number of cross sections per transition '
             '(default %(default)s)')

    # Output options
    output = parser.add_argument_group(
        'output options', description='Options controlling the generated '
        'output files')
    output.add_argument(
        '--2D', dest='output_2dm', action='store_true',
        help='generate 2DM mesh instead of 1D channel (default: %(default)s)')
    output.add_argument(
        '--BED', dest='bed', action='store_true',
        help='include a SOILDEF block in the output (default: %(default)s)')
    output.add_argument(
        '-BS', metavar='bed_start', dest='bed_start', default=3, type=int,
        help='starting index for channel bed definition [-] '
             '(default: %(default)s)')

    output.add_argument(
        '--R', '--reverse', dest='reverse', action='store_true',
        help='reverse cross section order (default: %(default)s)')
    output.add_argument(
        '-CS', metavar='prefix', dest='prefix', default='CS_', type=str,
        help='cross section name prefix (default: %(default)s)')

    # Parse user input
    args = parser.parse_args()

    # Set up channel factory
    factory: factories.AbstractChannelFactory
    if args.widening:
        factory = factories.TrapezoidalChannelWidening(
            dist_cs=args.dx,
            num_cs=args.num_cs,
            bed_width=args.bed_width,
            height=args.height,
            bed_slope=args.slope_channel,
            bank_slope=args.slope_banks,
            friction_bed=args.ks_bed,
            friction_bank=args.ks_banks,
            # Start of widening args
            widening_bed_width=args.w_bed_width,
            widening_friction=args.w_ks_bed,
            widening_dist_cs=args.w_dx,
            widening_num_cs=args.w_num_cs,
            transition_dist_cs=args.t_dx,
            transition_num_cs=args.t_num_cs,
            # End of widening args
            midpoint_shift=args.mp_shift,
            raise_by=args.offset_z)
    else:
        factory = factories.TrapezoidalChannel(
            dist_cs=args.dx,
            num_cs=args.num_cs,
            bed_width=args.bed_width,
            height=args.height,
            bed_slope=args.slope_channel,
            bank_slope=args.slope_banks,
            friction_bed=args.ks_bed,
            friction_bank=args.ks_banks,
            midpoint_shift=args.mp_shift,
            raise_by=args.offset_z)

    geometry = factory.build_channel()
    geometry.name = 'geometry'
    # Update cross sections
    print('Processing cross sections...')
    prefix = args.prefix
    for index, cs_ in enumerate(geometry.cross_sections):
        cs_.flow_axis_coord += args.offset_x
        cs_.name = f'{prefix}{index+1}'
        cs_.angle = 90.0

        # Define bottom range
        if args.bed:
            offset: int = args.bed_start
            vtx_count = len(cs_.vertices)
            cs_.bottom_slice = offset, vtx_count-offset
            cs_.soil_defs[1] = offset, vtx_count-offset

    if args.output_2dm:
        print('Writing 2D mesh geometry...')

        # Create MATID markers for all channel sections
        region_markers: List[RegionMarker] = []
        for index, cs_ in enumerate(geometry.cross_sections[:-1]):
            next_cs = geometry.cross_sections[index+1]
            dist = cs_.flow_axis_coord + (
                next_cs.flow_axis_coord - cs_.flow_axis_coord) * 0.5
            offsets_bank_r: List[float] = []
            for v_a, v_b in zip(cs_.vertices[4:7], next_cs.vertices[4:7]):
                offsets_bank_r.append((v_a.pos[0] + v_b.pos[0]) * 0.5)
            # Coordinates of the markers to create (right side only)
            marker_bank_bot = (offsets_bank_r[0] + offsets_bank_r[1]) * 0.5
            marker_bank_top = (offsets_bank_r[1] + offsets_bank_r[2]) * 0.5
            # Check whether this marker falls inside a widened cell
            this_bed_width = cs_.vertices[-3].pos[0] - cs_.vertices[2].pos[0]
            next_bed_width = (
                next_cs.vertices[-3].pos[0] - next_cs.vertices[2].pos[0])
            is_widened = (this_bed_width + next_bed_width) / 2 > args.bed_width
            # Region markers to insert
            region_markers.extend(
                (RegionMarker(dist, -marker_bank_top, attribute=3),
                 RegionMarker(dist, -marker_bank_bot, attribute=3),
                 RegionMarker(dist, 0.0, attribute=2 if is_widened else 1),
                 RegionMarker(dist, marker_bank_bot, attribute=3),
                 RegionMarker(dist, marker_bank_top,  attribute=3)))

        mesh = geometry.to_mesh(regions=region_markers)
        mesh.save(f'{prefix}geometry.2dm')

    else:
        print('Writing 1D mesh geometry...')
        with BaseChainWriter(f'{prefix}geometry.bmg') as writer:
            writer.write_channel(geometry)

        print('Writing BASEMENT command file...')
        # Reverse list if required
        cs_numbering = [i+1 for i in range(factory.cs_total)]
        if args.reverse:
            cs_numbering = list(reversed(cs_numbering))
        with open(f'{prefix}geometry.bmc', 'w') as cmd_file:
            cmd_file.write(
                '// BASEMENT v2.8 command file\n'
                f'// Generated 2020-11-10 09:27:56 using BASEchange {version}\n'
                'PROJECT {\n}\n'
                'DOMAIN {\n'
                '\tBASECHAIN_1D {\n'
                '\t\tGEOMETRY {\n'
                '\t\t\ttype = basement\n'
                f'\t\t\tfile = {prefix}geometry.bmg\n'
                f'\t\t\tcross_section_order = ('
                + ' '.join((f'{prefix}{i}' for i in cs_numbering))
                + ')\n'
                '\t\t}\n'
                '\t}\n'
                '}\n')



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Main wrapper class and utility code."""

import datetime
import os
import platform
import stat
import subprocess
from typing import (Any, Callable, Collection, Dict, Iterable, List, Optional,
                    Tuple)
from .log import logger
from .markers import HoleMarker, RegionMarker
from .pslg import PSLGNode, PSLGSegment
from .triangle_io import write_nodes, write_poly

# Path to the Triangle installation
_TRIANGLE_PATH = 'bin/'


class Triangle:
    """Wrapper class for Jonathan Shewchuk's "Triangle".

    This class is responsible for creating the input files, spawning
    the "Triangle" subprocesses, and eventually parsing the output.

    See the triangulate_nodes and triangulate_poly for usage
    instructions.

    Attributes
    ----------
    executable : str
        The executable chosen for the current platform, only provided
        for debug purposes
    """

    def __init__(self) -> None:
        """Initialize the Triangle namespace class.

        This mainly sets the internal executable for the current
        system platform.
        """
        logger.info('Initializing Triangle...')
        self.executable = self._find_executable()

    @staticmethod
    def _find_executable() -> str:
        """Retrieve the appropriate binary for the current system."""
        # Microsoft NT (Windows)
        if os.name == 'nt':
            filename = 'triangle_32.exe'
        # Unix
        elif os.name == 'posix':
            # Linux
            if '32bit' in platform.architecture():
                filename = 'triangle_linux_32'
            else:
                filename = 'triangle_linux_64'
        # Other (unsupported)
        else:
            raise RuntimeError(f'Unsupported operating system "{os.name}"')

        # Look for the appropriate executable
        folder = os.path.dirname(__file__)
        triangle_path = os.path.join(folder, _TRIANGLE_PATH, filename)
        if not os.path.isfile(triangle_path):
            raise RuntimeError(
                f'Unable to locate executable "{triangle_path}"')
        # Set executable flag for unix systems
        if os.name == 'posix':
            os.chmod(triangle_path, stat.S_IEXEC)
        logger.info('Compatible Triangle binary found: %s', filename)
        return triangle_path

    @staticmethod
    def _get_base_name() -> str:
        """Return a default triangle input filename.

        This name will generally not be visible to the user and is
        mainly intended to be unique for debug purposes.
        """
        timestamp = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
        return f'triangle_{timestamp}'

    @staticmethod
    def _get_output_names(
            filename: str, *, node_file: bool = True, ele_file: bool = True,
            poly_file: bool = True, edge_file: bool = False,
            neigh_file: bool = False) -> Tuple[str, ...]:
        """Return the output filenames for a given input filename.

        This function will take in a given input filename or path and
        return the expected output filename generated by Triangle.

        Parameters
        ----------
        filename : str
            The input file passed to Triangle
        node_file : bool, optional
            Whether to include NODE output, by default True
        ele_file : bool, optional
            Whether to include ELE output, by default True
        poly_file : bool, optional
            Whether to include POLY output, by default True
        edge_file : bool, optional
            Whether to include EDGE output, by default False
        neigh_file : bool, optional
            Whether to include NEIGH output, by default False

        Returns
        -------
        Tuple[str, ...]
            The output file names that will be generated by Triangle
        """
        # Strip off any path information
        base_name = os.path.basename(filename)
        # Right-split the base name up to two times, one for the iteration
        # number, the other for the file extension
        items = base_name.rsplit('.', 2)
        iter_num = 0
        if len(items) >= 3:
            base_name = items[0]
            # If the second word is an int, treat it as an iteration number,
            # otherwise assume it's part of the filename.
            try:
                iter_num = int(items[1])
            except ValueError:
                base_name += '.' + items[1]
        else:
            base_name = items[0]
        # Increment iteration number
        iter_num += 1
        # Determine the output extensions to include
        extensions = {'node': node_file, 'ele': ele_file, 'poly': poly_file,
                      'edge': edge_file, 'neigh': neigh_file}
        output_files: Tuple[str, ...] = tuple(
            f'{base_name}.{iter_num}.{k}' for k, v in extensions.items() if v)
        return output_files

    @staticmethod
    def _move_to_current(dir_: str, *args: str) -> None:
        """Move any files in *args to the current working directory.

        Parameters
        ----------
        dir_ : str
            The directory containing the input files
        *args: str
            The filenames to move to the CWD
        """
        for filename in args:
            temp_name = os.path.join(dir_, filename)
            os.replace(temp_name, filename)

    def triangulate_nodes(self, nodes: Collection[PSLGNode],
                          segments: Optional[Collection[PSLGSegment]] = None,
                          node_attr_keys: Optional[List[str]] = None,
                          *, dir_: str = '', **kwargs: Any) -> Tuple[str, str]:
        """Triangulate the given input nodes.

        This effectively creates an elevation mesh of the input points.

        Mesh boundary polygons must be split into lines and included in
        the segments argument.

        Any extra keyword arguments will be passed on to the run()
        method.

        Parameters
        ----------
        nodes : Collection[PSLGNode]
            The input nodes to process
        dir_ : str, optional
            The directory to run Triangle in, by default '.'
        segments : Collection[PSLGSegment], optional
            Any input segments to enforce during triangulation, by
            default None
        node_attr_keys : List[str], optional
            The list of node attributes to use during triangulation, by
            default None
        **kwargs: Any
            Additional keyword arguments to pass to the run() method

        Returns
        -------
        Tuple[str, str]
            The Triangle output files
        """
        base_name = os.path.join(dir_, self._get_base_name())

        ext = 'node' if segments is None else 'poly'
        input_name = f'{base_name}.{ext}'
        input_file = os.path.join(dir_, input_name)
        if segments is None:
            write_nodes(input_file, nodes, node_attr_keys=node_attr_keys)
        else:
            write_poly(input_file, nodes=nodes, segments=segments,
                       node_attr_keys=node_attr_keys)
        try:
            self.run(input_file, input_is_poly=segments is not None,
                     no_output_poly=True, redirect_stdout=_inline_print,
                     **kwargs)
        except subprocess.CalledProcessError as err:
            print('An error occurred during the execution of Triangle')
            raise err
        node_file, ele_file = self._get_output_names(
            base_name, poly_file=False)
        return os.path.join(dir_, node_file), os.path.join(dir_, ele_file)

    def triangulate_poly(self, nodes: Collection[PSLGNode],
                         segments: Optional[Collection[PSLGSegment]] = None,
                         holes: Optional[Iterable[HoleMarker]] = None,
                         regions: Optional[Iterable[RegionMarker]] = None,
                         *, dir_: str = '',
                         node_attr_keys: Optional[List[str]] = None,
                         use_matid: bool = True, **kwargs: Any
                         ) -> Tuple[str, str]:
        """Triangulate the given input data.

        This supports all the functionality of triangulate_nodes, but
        also provides access to segment definitions, holes and region
        markers.

        Parameters
        ----------
        nodes : Collection[PSLGNode]
            Input nodes to triangulate
        dir_ : str, optional
            The directory to run Triangle in, by default '.'
        segments : Collection[PSLGSegment], optional
            Input segments to enforce during triangulation, by default
            None
        holes : Iterable[HoleMarker], optional
            Optional hole markers, by default None
        regions : Iterable[RegionMarker], optional
            Optionaln region markers, by default None
        node_attr_keys : List[str], optional
            The node attribute keys to use during triangulation, by
            default None
        use_matid : bool, optional
            Whether to include material IDs, by default True

        Returns
        -------
        Tuple[str, str]
            Triangle output files
        """
        base_name = self._get_base_name()

        input_name = f'{base_name}.poly'
        input_file = os.path.join(dir_, input_name)
        extra_args: Dict[str, Any] = {}
        if segments is not None:
            extra_args['segments'] = segments
        if holes is not None:
            extra_args['holes'] = holes
        reg_area = False
        if regions:
            extra_args['regions'] = regions
            reg_area = True
        write_poly(input_file, nodes=nodes, node_attr_keys=node_attr_keys,
                   include_matid=use_matid, **extra_args)
        try:
            self.run(input_file, input_is_poly=True,
                     no_output_poly=True, redirect_stdout=_inline_print,
                     use_region_attributes=use_matid,
                     use_region_areas=reg_area,
                     **kwargs)
        except subprocess.CalledProcessError as err:
            print('An error occurred during the execution of Triangle.')
            raise err
        node_file, ele_file = self._get_output_names(
            base_name, poly_file=False)
        return os.path.join(dir_, node_file), os.path.join(dir_, ele_file)

    def run(self, input_file: str, *, suppress_output: bool = False,
            redirect_stdout: Optional[Callable[[str], None]] = None,
            flags: Optional[Dict[str, Any]] = None, **kwargs: Any) -> None:
        """Run Triangle using the provided input parameters.

        This is the most low-level interface for interfacing with
        "Triangle". Note that this file does not automatically generate
        its own input files.

        Parameters
        ----------
        input_file : str
            The input POLY or NODE file to run through Triangle
        suppress_output : bool, optional
            Whether to suppress all Triangle output during execution,
            by default False
        redirect_stdout : Callable[[str], None], optional
            A callable to pass any Triangle output to, rather than
            printing it, by default None
        flags : Dict[str, Any], optional
            Custom Triangle command line flags; flags defined here will
            overwrite those defined by other sources, by default None

        Raises
        ------
        ValueError
            Raised if the input file is neither a NODE nor POLY file
        subprocess.CalledProcessError
            Raised if Triangle returns a non-zero exit code
        """
        # Validate input filename
        if os.path.splitext(input_file)[1].lower() not in ('.poly', '.node'):
            raise ValueError('Input file must be a NODE or POLY file')
        # Extract custom triangle arguments
        custom_args = kwargs.pop('custom_args', '')
        # Generate the arguments and command string for Triangle execution
        flag_dict = dict(kwargs)
        if flags is not None:
            flag_dict.update(flags)
        flag_str = self._parse_flags(**flag_dict)
        command = f'{self.executable} -{flag_str}{custom_args} {input_file}'
        # Instantiate a new subprocess
        logger.info('Running Triangle: %s', command)
        kwargs = {}
        if os.name == 'nt':
            kwargs.update({'creationflags': subprocess.CREATE_NO_WINDOW})
        else:
            kwargs.update({'shell': True})
        process = subprocess.Popen(command, stdout=subprocess.PIPE,
                                   universal_newlines=True, **kwargs)
        logger.debug('Triangle subprocess spawned (PID %d)', process.pid)
        try:
            # Process subprocess output as it comes in
            for line in iter(process.stdout.readline, ''):  # type: ignore
                line = str(line)
                logger.debug('Subprocess: %s', line.rstrip())
                if suppress_output:
                    continue
                if redirect_stdout:
                    redirect_stdout(line)
        finally:
            process.stdout.close()  # type: ignore
        # Wait for subprocess completion and raise appropriate errors
        return_code = process.wait()
        logger.debug(
            'Triangle subprocess exited with exit status %d', return_code)
        if return_code:
            raise subprocess.CalledProcessError(return_code, command)

    @staticmethod
    def _parse_flags(**kwargs: str) -> str:
        """Convert input keyword arguments to "Triangle" flags.

        This method parses any number of input kwargs and returns the
        appropriate Triangle comand string.

        Returns
        -------
        str
            The Triangle command string

        Raises
        ------
        ValueError
            Raised if unexpected keys are found in the input
        """
        # The following dict maps the argument-less command line switches to
        # simplify the elif chain below.
        switches = {
            'input_is_poly': 'p',
            'conforming_delaunay': 'D',
            'jettison_nodes': 'j',
            'write_edges': 'e',
            'write_neigh': 'n',
            'no_boundary_in_output': 'B',
            'no_iteration_numbers': 'I',
            'ignore_holes': 'O',
            'no_steiner_on_boundary': 'Y',
            'no_steiner_on_segment': 'YY',
            'quiet': 'Q',
            'verbose': 'V',
            'keep_convex_hull': 'c',
            # 'refine': 'r',
            'no_output_node': 'N',
            'no_output_ele': 'E',
            'no_output_poly': 'P',
            'use_region_areas': 'a',
            'use_region_attributes': 'A'
        }
        flags = ''
        for key, value in kwargs.items():
            # Basic switches
            if key in switches:
                if value:
                    flags += switches[key]
                continue
            # Switches with arguments
            if key == 'min_angle':
                flags += f'q{value}'
            elif key == 'max_area':
                flags += f'a{value}'
            elif key == 'max_steiner':
                flags += f'S{value}'
            elif key == 'debug_level':
                flags += 'V' * int(value)
            # Fallback
            else:
                raise ValueError(f'Unknown triangle flag: {key}={value}')
        return flags


def _inline_print(string: str) -> None:
    """Like the default print, but without automatic newlines.

    Parameters
    ----------
    string : str
        The string to print
    """
    print(string, end='')



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Mesh generation utility."""

import os
import shutil
import tempfile
from typing import Any, Collection, Dict, Iterable, List, Optional
from . import triangle
from .core import Mesh, MeshNode
from .log import logger

# Name for the node attribute used during elevation meshing via Triangle
_Z_ATTR = 'pos_z'


def elevation_mesh(nodes: Collection[triangle.PSLGNode],
                   segments: Optional[Collection[triangle.PSLGSegment]] = None,
                   *, move_tempfiles: Optional[str] = None,
                   **kwargs: Any) -> Mesh:
    """Triangulate the given input vertices and break lines.

     Generate a mesh using the given vertices. No mesh quality
    constraints will be applied. The output mesh will contain elevation
    information.

    Any excess kwargs are passed on to the Triangle.run() method.

    Parameters
    ----------
    nodes : Collection[triangle.PSLGNode]
        The set of nodes to mesh
    segments : Collection[triangle.PSLGSegment], optional
        Optional line segments connecting the nodes, by default None
    move_tempfiles : str, optional
        A directory. If provided, the temporary Triangle input and
        result files are moved to this directory instead of being
        deleted, by default None

    Returns
    -------
    Mesh
        Triangulated mesh output
    """
    logger.info('Entering elevation meshing utility')

    # Create a temporary directory to run Triangle in
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir = str(temp_dir)
        logger.debug('Created temporary working directory: %s', temp_dir)
        # Run triangle
        results = triangle.Triangle().triangulate_nodes(
            nodes, segments=segments, dir_=temp_dir,
            node_attr_keys=[_Z_ATTR], **kwargs)
        # Create a mesh from the triangle result files
        mesh = _mesh_from_triangle(*results, node_attrs=[_Z_ATTR],
                                   use_matid=False,
                                   delete_input=not move_tempfiles)
        # Optionally move the tempfiles out of the temporary directory before
        # its destruction
        if move_tempfiles is not None:
            _move_tempfiles(temp_dir, move_tempfiles)
    logger.debug('Temporary working directory abandoned')

    # Take the elevation information from the node attribute and move the node
    # to its designated elevation. This converts the 2D Triangle result mesh
    # into a 2.5D elevation mesh.
    for node in mesh.nodes:
        node.move(z=float(node.attributes.pop(_Z_ATTR)))
    return mesh


def quality_mesh(nodes: Collection[triangle.PSLGNode],
                 segments: Optional[Collection[triangle.PSLGSegment]] = None,
                 holes: Optional[Iterable[triangle.HoleMarker]] = None,
                 regions: Optional[Iterable[triangle.RegionMarker]] = None, *,
                 conforming_delaunay: bool = True,
                 preserve_boundary: bool = False,
                 min_angle: Optional[float] = None,
                 max_area: Optional[float] = None,
                 move_tempfiles: Optional[str] = None,
                 **kwargs: Any) -> Mesh:
    """Advanced Meshing with mesh quality constraints applied.

    Any excess kwargs are passed on to the Triangle.run() method.

    Parameters
    ----------
    nodes : Collection[triangle.PSLGNode]
        The nodes to mesh
    segments : Collection[triangle.PSLGSegment], optional
        Optional line segments connecting the given nodse, by default
        None
    holes : Iterable[triangle.HoleMarker], optional
        Markers flagging a line-bounded region as a hole, by default
        None
    regions : Iterable[triangle.RegionMarker], optional
        Markers used to  assign attributes to their respective
        line-bounded region, by default None
    conforming_delaunay : bool, optional
        Whether to use conforming delaunay instead of contrained
        delaunay, by default True
    preserve_boundary : bool, optional
        If true, no nodes will be inserted along the mesh boundary, by
        default False
    min_angle : float, optional
        Minimum angle contraints for mesh elements, by default None
    max_area : float, optional
        Global maximum area constraint for mesh elements, by default
        None
    move_tempfiles : str, optional
        A directory. If provided, the temporary Triangle input and
        result files are moved to this directory instead of being
        deleted, by default None

    Returns
    -------
    Mesh
        The triangulated quality mesh

    Raises
    ------
    ValueError
        Raised if no breaklines or input vertices could be inferred
    """
    logger.info('Entering quality meshing utility')

    # Set up triangle flags
    flags: Dict[str, Any] = {'conforming_delaunay': conforming_delaunay,
                             'no_steiner_on_boundary': preserve_boundary}
    if max_area is not None:
        flags['max_area'] = max_area
    if min_angle is not None:
        flags['min_angle'] = min_angle
    flags.update(kwargs)
    # Ensure at least some break lines have been provided
    if not segments:
        raise ValueError('No breaklines or input vertices provided')

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir = str(temp_dir)
        logger.debug('Created temporary working directory: %s', temp_dir)
        # Run triangle
        results = triangle.Triangle().triangulate_poly(
            nodes, segments, holes, regions,
            use_matid=bool(regions), dir_=temp_dir, **flags)
        # Create mesh from triangle result files
        mesh = _mesh_from_triangle(*results, use_matid=bool(regions),
                                   delete_input=not move_tempfiles)
        # Optionally move the tempfiles out of the temporary directory before
        # its destruction
        if move_tempfiles is not None:
            _move_tempfiles(temp_dir, move_tempfiles)
    logger.debug('Temporary working directory abandoned')
    return mesh


def _mesh_from_triangle(
        node_file: str, ele_file: str, node_attrs: Optional[List[str]] = None,
        use_matid: bool = True, delete_input: bool = False) -> Mesh:
    """Create a new mesh from Triangle output files.

    Note that the output mesh will not have any elevation information
    associated with its nodes whatsoever. All nodes will have the
    elevation 0.0.

    Parameters
    ----------
    node_file : str
        Path to the Triangle NODE file
    ele_file : str
        Path to the Triangle ELE file
    node_attrs : List[str], optional
        Keys used to match the positional Triangle attributes with
        named attributes, by default None
    use_matid : bool, optional
        Whether to read MATIDs from file, by default True
    delete_input : bool, optional
        Whether to dlete the input files after completion, by default
        False

    Returns
    -------
    Mesh
        A mesh generated from the Triangle output files
    """
    # Read triangle input
    pslg_nodes = triangle.triangle_io.read_nodes(node_file, node_attrs)
    pslg_elements = triangle.triangle_io.read_elements(ele_file, pslg_nodes,
                                                       use_matid=use_matid)

    nodes: Dict[int, MeshNode] = {}
    mesh = Mesh()
    for pslg_ele in pslg_elements:
        # Add element nodes to mesh
        for pslg_node in pslg_ele.nodes:
            assert pslg_node.id_ is not None
            if pslg_node.id_ in nodes.keys():
                continue
            point = (*pslg_node.as_tuple(), 0.0)
            node = mesh.add_node(point, id_=pslg_node.id_,
                                 **pslg_node.attributes)
            nodes[pslg_node.id_] = node
        # Add element
        element_nodes = tuple(
            nodes[n.id_] for n in pslg_ele.nodes)  # type: ignore
        matids = () if not use_matid else (
            (pslg_ele.matid,) if pslg_ele.matid is not None else ())
        mesh.add_element(element_nodes, pslg_ele.id_, *matids)  # type: ignore
    if delete_input:
        os.remove(node_file)
        os.remove(ele_file)
    return mesh


def _move_tempfiles(source_dir: str, target_dir: str) -> None:
    """Move the temporary Triangle files to another directory.

    All Triangle operations are performed in a temporary directory that
    is deleted after meshing. This methods moves them somewhere else
    for instrospection and debugging.

    Parameters
    ----------
    source_dir : str
        The source directory to copy the files from
    target_dir : str
        The target directory to copy the files to
    """
    logger.debug('Moving temporary Triangle files to %s', target_dir)
    for filename in os.listdir(source_dir):
        file_path = os.path.join(source_dir, filename)
        shutil.move(file_path, target_dir)



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Implementation of the internal mesh representation for BASEmesh."""

import os
import warnings
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union
import numpy as np  # type: ignore
from ..abc import ElevationSource, SpatialCollection, SupportsSpatial
from ..algorithms import (counting_sort, dist_2d, interpolate_triangle,
                          point_in_polygon_concave, point_in_polygon_convex,
                          distance_to_polygon)
from ..containers import SpatialSet
from ..log import logger
from ..types import Point2D, Point3D, Polygon2D, Triangle2D, Triangle3D
from .geometry import Node

try:
    import py2dm
except ModuleNotFoundError:
    import sys
    wheel_dir = os.path.dirname(__file__)
    wheel_dir, _ = os.path.split(wheel_dir)
    whl_typing_extensions = os.path.join(
        wheel_dir, 'packages', 'typing_extensions-3.7.4.3-py3-none-any.whl')
    sys.path.append(whl_typing_extensions)
    whl_py2dm = os.path.join(
        wheel_dir, 'packages', 'py2dm-0.1.0-py3-none-any.whl')
    sys.path.append(whl_py2dm)

    import py2dm


# Type aliases
_NodeOrPt = Union[Tuple['MeshNode', 'MeshNode', 'MeshNode'],
                  Tuple[Point3D, Point3D, Point3D]]


class MeshNode(Node, SupportsSpatial):
    """A node in a mesh.

    This subclasses the Node class and extends it by providing support
    for elements, a reference to the containing mesh, as well as a
    unique node ID.
    """

    __slots__ = ('_attached_elements', '_id', '_mesh')

    def __init__(self, mesh: 'Mesh', point: Point3D,
                 id_: int, **kwargs: Any) -> None:
        """Initialise a new mesh node.

        Note that the preferred means of creating mesh nodes is the
        Mesh.add_node() method.

        Any keyword arguments passed will be stored with the node and
        are accessible through its attributes property.

        Parameters
        ----------
        mesh : Mesh
            The mesh to create the node in
        point : Tuple[float, float, float]
            The location of the node in space
        id_ : int
            The unique ID of the mesh node
        """
        super().__init__(point, **kwargs)  # Initialise Node
        self._mesh = mesh
        self.id_ = id_
        self._attached_elements: Set['MeshElement'] = set()

    @property
    def elements(self) -> Set['MeshElement']:
        """Return a set of mesh elements defined using this node.

        These elements in this set are also the ones that will be
        removed if this node is deleted.
        """
        return self._attached_elements

    @property
    def mesh(self) -> 'Mesh':
        """Return the mesh this node is a part of."""
        return self._mesh

    @property
    def spatial_marker(self) -> Point2D:
        """Return a 2D point representing this object.

        This point will be used by space-aware containers to optimise
        memory layout.
        """
        return self.pos_x, self.pos_y

    def attach_element(self, element: 'MeshElement') -> None:
        """Attach a mesh element to the mesh node.

        This is usually done by the MeshElement initialiser to ensure
        the mesh element can be deleted when its defining node is.

        Parameters
        ----------
        element : MeshElement
            The element to attach
        """
        self._attached_elements.add(element)

    def detach_element(self, element: 'MeshElement') -> None:
        """Detach a mesh element from the mesh node.

        This is used to remove all references to the mesh element upon
        its deletion, allowing the garbage collector to dispose of it.

        Parameters
        ----------
        element : MeshElement
            The element to detach
        """
        self._attached_elements.discard(element)


class MeshElement(SupportsSpatial):
    """A single mesh element, or face, in a mesh."""

    __slots__ = ('id_', 'materials', '_mesh', '_node_1', '_node_2', '_node_3')

    def __init__(self, mesh: 'Mesh',
                 nodes: Tuple[MeshNode, MeshNode, MeshNode],
                 id_: int, *materials: Union[int, float]) -> None:
        """Initialise a new mesh element.

        Note that the preferred means of creating new mesh elements is
        using the Mesh.add_element() method.

        The element will automatically oriented, you do not have to use
        a specific order when passing the mesh nodes.

        Parameters
        ----------
        mesh : Mesh
            The mesh to create this element for
        nodes : Tuple[MeshNode, MeshNode, MeshNode]
            Existing mesh nodes to create the element from
        id_ : int
            The unique ID of the mesh element
        *materials : Union[int, float]
            Any number of material indices for the mesh
        """
        self.id_ = id_
        self.materials = tuple(materials)
        self._mesh = mesh
        self._node_1 = nodes[0]
        self._node_2 = nodes[1]
        self._node_3 = nodes[2]
        nodes[0].attach_element(self)
        nodes[1].attach_element(self)
        nodes[2].attach_element(self)

        # Conform the element
        # NOTE: calling point_in_polygon_convex for an inner point of the
        # triangle will only return True if the triangle uses CCW vertex order.
        if not point_in_polygon_convex(
                self.spatial_marker, self.as_triangle_2d):
            # Swapping two nodes effectively flips the entire element
            self._node_1 = nodes[1]
            self._node_2 = nodes[0]

    @property
    def area(self) -> float:
        """Return the area of the mesh element."""
        # Convert to numpy arrays
        pt_a = np.array(self.points[0])
        pt_b = np.array(self.points[1])
        pt_c = np.array(self.points[2])
        # Get vectors
        vec_ab = pt_b - pt_a
        vec_ac = pt_c - pt_a
        # Get cross product and halve it
        return float(np.cross(vec_ab, vec_ac)) / 2.0

    @property
    def center(self) -> Tuple[float, float, float]:
        """Return the center of mass of the mesh element."""
        avg_x = (self.points[0][0] + self.points[1][0] + self.points[2][0]) / 3
        avg_y = (self.points[0][1] + self.points[1][1] + self.points[2][1]) / 3
        avg_z = (self.points[0][2] + self.points[1][2] + self.points[2][2]) / 3
        return avg_x, avg_y, avg_z

    @property
    def mesh(self) -> 'Mesh':
        """Return the mesh this element is a part of."""
        return self._mesh

    @property
    def nodes(self) -> Tuple[MeshNode, MeshNode, MeshNode]:
        """Return the nodes defining this element."""
        return self._node_1, self._node_2, self._node_3

    @property
    def points(self) -> Triangle3D:
        """Return the tuple of points defining the mesh element."""
        return (self._node_1.as_tuple(), self._node_2.as_tuple(),
                self._node_3.as_tuple())

    @property
    def as_triangle_2d(self) -> Triangle2D:
        """Return the tuple of 2D points representing this element."""
        return (self._node_1.spatial_marker, self._node_2.spatial_marker,
                self._node_3.spatial_marker)

    @property
    def spatial_marker(self) -> Point2D:
        """Return a 2D point representing this object.

        This point will be used by space-aware containers to optimise
        memory layout.
        """
        return self.center[:2]

    def contains(self, point: Union[Point2D, Point3D]) -> bool:
        """Check whether the given point lies within the element.

        If a 3D point was passed, only the first two coordinates will
        be used for containment checking.

        Parameters
        ----------
        point : Union[Tuple[float, float], Tuple[float, float, float]]
            The point to check for containment

        Returns
        -------
        bool
            Whether the point lies within the mesh element
        """
        return point_in_polygon_convex(point[:2], self.as_triangle_2d)

    def interpolate(self, point: Tuple[float, ...]) -> float:
        """Return the interpolated height at a given point.

        Note that the point does not have to lie within the mesh
        element for this. The element's nodes are used to define an
        infinite plane onto which the point is projected.

        If a 3D point was passed, its elevation will be ignored. It is
        only available as an input for convenience.

        Parameters
        ----------
        point : Union[Tuple[float, float], Tuple[float, float, float]]
            The point to interpolate

        Returns
        -------
        float
            The elevation of the input point when projected onto the
            element
        """
        return interpolate_triangle((point[0], point[1]), self.points)


class Mesh(ElevationSource):
    """A two dimensional mesh with node elevations (aka. 2.5D)."""

    __slots__ = ('_elements', '_nodes', 'node_strings')

    def __init__(self) -> None:
        """Initialise a new mesh.

        A mesh instantiated manually is always empty. Use the
        add_node() and add_element() methods to add geometry.
        """
        self._nodes: SpatialSet[MeshNode] = SpatialSet()
        self._elements: SpatialSet[MeshElement] = SpatialSet()
        self.node_strings: Dict[str, List[MeshNode]] = {}

    @classmethod
    def open(cls, filename: str) -> 'Mesh':
        """Read a 2DM mesh file and instantiate a mesh.

        This only honours the "ND" and "E3T" tags, with only up to one
        material ID.

        Returns
        -------
        Mesh
            The mesh as parsed from the 2DM file

        Raises
        ------
        ValueError
            Raised if the given file is not a valid 2DM file
        ValueError
            Raised if a negative node ID is encountered
        ValueError
            Raised if a negative element ID is encountered
        """
        nodes = []
        with py2dm.Reader(filename) as file_:
            mesh = cls()
            for node in file_.iter_nodes():
                mesh_node = mesh.add_node(node.pos, id_=node.id)
                nodes.append(mesh_node)
            for element in file_.iter_elements():
                element_nodes = tuple(nodes[n-1] for n in element.nodes)
                materials = element.materials
                mesh.add_element(
                    element_nodes, element.id, *materials)  # type: ignore
            for node_string in file_.iter_node_strings():
                node_string_nodes = [nodes[n-1] for n in node_string.nodes]
                if node_string.name is None:
                    warnings.warn(f'Node string {node_string} has no name, '
                                  'skipping')
                mesh.add_node_string(
                    node_string.name, node_string_nodes)   # type: ignore
        return mesh

    @property
    def elements(self) -> SpatialCollection[MeshElement]:
        """Return the mesh elements making up the mesh."""
        return self._elements

    @property
    def nodes(self) -> SpatialCollection[MeshNode]:
        """Return the mesh nodes defining its elements.

        Note that the mesh nodes are also available through the mesh
        elements via MeshElement.nodes, which may be faster than
        manually associating nodes and elements.
        """
        return self._nodes

    def add_node(self, point: Point3D, id_: Optional[int] = None,
                 **kwargs: Any) -> MeshNode:
        """Create a new node for the mesh.

        If no mesh ID was provided, a new mesh ID will be chosen based
        on existing nodes in the mesh.

        Parameters
        ----------
        point : Tuple[float, float, float]
            The position of the new node
        id_ : int, optional
            The ID to use for this node, by default None

        Returns
        -------
        MeshNode
            The node that was created; the node will already be added
            to the mesh by the time this method exits
        """
        if id_ is None:
            id_ = self._get_node_id()
        node = MeshNode(self, point, id_, **kwargs)
        self._nodes.add(node)
        return node

    def add_element(self,
                    nodes: Union[Tuple['MeshNode', 'MeshNode', 'MeshNode'],
                                 Tuple[Point3D, Point3D, Point3D]],
                    id_: Optional[int], *materials: Union[int, float]
                    ) -> MeshElement:
        """Add a new element using the given mesh nodes.

        If the given nodes are already part of the current mesh, they
        will be linked to the created element. If they are points, new
        nodes will be created and added to the mesh before adding the
        element.

        If no element ID has been provided, one will be selected based
        on existing elements.

        Parameters
        ----------
        nodes : Union[Tuple[MeshNode, MeshNode, MeshNode],
                      Tuple[Point3D, Point3D, Point3D]]
            The nodes to create the element from
        id_ : int, optional
            The ID of the element, by default None
        *materials : int, optional
            Any number of material indices for the element

        Returns
        -------
        MeshElement
            The mesh element that was created

        Raises
        ------
        ValueError
            Raised when attempting to define an element using nodes
            defined in another mesh
        ValueError
            Raised if any of the input nodes coincide
        """
        # Process input nodes
        element_nodes: Set[MeshNode] = set()
        for node in nodes:
            if isinstance(node, MeshNode):
                # If the node is part of the mesh, re-use it
                if node in self.nodes:
                    element_nodes.add(node)
                # If it isn't, raise an error (no cross-meshing yet)
                else:
                    raise ValueError(
                        f'MeshNode {node} is not part of the mesh')
            else:
                # If the node is a tuple of floats, look for an existing node
                # with the same coordinates
                try:
                    found_node = self.node_at(node)
                except ValueError:
                    # If it doesn't exist yet, create a new node
                    new_node = MeshNode(self, node, self._get_node_id())
                    element_nodes.add(new_node)
                else:
                    # If it does, add the existing mesh node instead
                    element_nodes.add(found_node)

        # Ensure our three nodes span a triangle (i.e. don't coincide)
        if len(set(n.spatial_marker for n in element_nodes)) < 3:
            raise ValueError('Input nodes may not coincide')

        # Add the nodes to the mesh, if required
        for node in element_nodes:
            if node not in self.nodes:
                self._nodes.add(node)

        # Create a new element
        if id_ is None:
            id_ = self._get_element_id()
        nodes_tuple: Tuple[MeshNode, MeshNode, MeshNode] = tuple(
            element_nodes)  # type: ignore
        element = MeshElement(self, nodes_tuple, id_, *materials)
        self._elements.add(element)
        return element

    def add_node_string(self, name: str, nodes: Iterable[MeshNode]) -> None:
        """Add a new node string using the given mesh nodes.

        If the given nodes are not part of the current mesh, an error
        will be raised.

        Parameters
        ----------
        name: str
            The unique name of the node string
        nodes: Iterable[MeshNode]
            The series of nodes making up the node string

        Raises
        ------
        KeyError
            Raised if a node string with the given name already exists
        ValueError
            Raised if the given nodes are not part of the current mesh
        """
        if name in self.node_strings:
            raise KeyError(f'Node string {name} already exists')
        for node in nodes:
            if node.mesh != self:
                raise ValueError(
                    f'Node {node} is not part of the current mesh')
        self.node_strings[name] = list(nodes)

    def element_at(self, point: Union[Point2D, Point3D]) -> MeshElement:
        """Return the element containing the given point.

        If the point lies outside of any mesh elements, a ValueError
        is raised.

        If a 3D point is passed, only the first two coordinates will be
        used for containment checking.

        Parameters
        ----------
        point : Union[Tuple[float, float], Tuple[float, float, float]]
            The point to find the containing element for

        Returns
        -------
        MeshElement
            The mesh element containing the given point

        Raises
        ------
        ValueError
            Raised if the input point does not lie within any mesh
            element
        """
        for element in self.elements.iter_spatial(point[:2]):
            if element.contains(point):
                return element
        raise ValueError(f'Point {point} does not lie within any mesh element')

    def elements_by_polygon(self, polygon: Polygon2D,
                            check_midpoints: bool = True) -> List[MeshElement]:
        """Return any elements contained in the given polygon.

        Disabling the "check_midpoints" flag will cause every corner
        point of the element to be checked instead, which will triple
        the workload.

        Parameters
        ----------
        polygon : Tuple[Tuple[float, float], ...]
            The polygon to filter the elements by
        check_midpoints : bool, optional
            Whether to use the (faster) midpoint-only check, or the
            more accurate edge containment check, by default True

        Returns
        -------
        List[MeshElement]
            A list of mesh elements contained in the polyogn

        Raises
        ------
        ValueError
            Raised if the input polygon has fewer than three vertices
        """
        # Simple check: centroids only
        if check_midpoints:
            return [e for e in self.elements
                    if point_in_polygon_concave(e.center[:2], polygon)]
        # Advanced check: all element vertices must be contained
        return [e for e in self.elements
                if all(point_in_polygon_concave(n.as_tuple_2d(), polygon)
                       for n in e.nodes)]

    def elevation_at(self, point: Union[Point2D, Point3D]) -> float:
        """Return the interpolated elevation at the given point.

        If no mesh element contains the point exactly, it will be
        re-checked using a maximum distance of 1.0e-6.

        If a 3D point is passed, only the first two coordinates will be
        used for containment checking.

        Parameters
        ----------
        point : Union[Tuple[float, float], Tuple[float, float, float]]
            The point to get the elevation for

        Returns
        -------
        float
            The elevation of the mesh at the given point.

        Raises
        ------
        ValueError
            Raised if the point lies outside the mesh
        """
        PRECISION = 1.0e-1  # Maximum distance for non-exact point matches
        try:
            # Try to find an element containing this point
            element = self.element_at(point)
        except ValueError as err:
            # If this fails, try to find the closest element instead. This is
            # necessary as the element_at check fails early and does not check
            # for floating point errors around mesh element edges.
            logger.debug('Unable to find containing element for point %s, '
                         'searching for closest element instead...', point[:2])
            element = self.get_element(point)

            # Calculate the minimum distance between the closest element and
            # the point to check
            triangle_2d = tuple(p[:2] for p in element.points)
            dist = distance_to_polygon(point[:2], triangle_2d)

            if dist < PRECISION:
                # If the point is within the precision tolerance of the
                # element, the element may still be used for interpolation.
                logger.debug('Closest element %s within tolerance (%f <= %f)',
                             element.spatial_marker, dist, PRECISION)
            else:
                # The point is too far away from any element,
                # re-raise the error and try another interpolation source.
                logger.debug('Closest element %s outside of tolerance (%f > '
                             '%f)', element.spatial_marker, dist, PRECISION)
                raise err
        return element.interpolate(point)

    def get_element(self, point: Union[Point2D, Point3D], *,
                    search_effort: float = 1.0) -> MeshElement:
        """Return the element closest to the given point.

        By default, this exhaustively searches the entire mesh. In this
        case, the search is guaranteed to return the closest element.

        The search_effort argument may be used to reduce search effort
        by quitting after X ratio of all elements have been checked,
        then picking the closest one found.
        This is only useful with space-aware containers where only
        searching ~10% of the entire data set might give sufficiently
        high confidence.

        If a 3D point is passed, only the first two coordinates will be
        used for containment checking.

        Parameters
        ----------
        point : Union[Tuple[float, float], Tuple[float, float, float]]
            The point to check
        search_effort : float, optional
            The ratio of the data set to check, by default 1.0

        Returns
        -------
        MeshElement
            The closest element to the given point

        Raises
        ------
        ValueError
            Raised if the search_effort argument valls is equal to or
            less than 0, or if it is greater than 1
        ValueError
            Raised if the mesh is empty
        """
        if search_effort <= 0.0 or search_effort > 1.0:
            raise ValueError('Search effort must fall within the (0.0, 1.0] '
                             f'range (current: {search_effort}')
        if not self.elements:
            raise ValueError('Mesh does not contain any element')

        point_2d: Point2D = point[:2]
        lowest_dist = -1.0
        closest_element: Optional[MeshElement] = None

        for index, element in enumerate(self.elements.iter_spatial(point_2d)):
            # Calculate the distance between the element and the point
            triangle_2d = tuple((p[0], p[1]) for p in element.points)
            dist = distance_to_polygon(point_2d, triangle_2d)

            # Exit early if the point is contained within an element
            if dist < 0.0:
                logger.info('Found containing element %s for point %s',
                            element.spatial_marker, point_2d)
                return element

            # Update closest element
            if lowest_dist < 0.0 or dist < lowest_dist:
                lowest_dist = dist
                closest_element = element

            # Check search effort
            if (index+1) / len(self.elements) > search_effort:
                logger.info(
                    'Exceeded search effort of %f: %d of %d elements checked',
                    search_effort, index+1, len(self.elements))
                break

        assert closest_element is not None
        return closest_element

    def _get_element_id(self) -> int:
        """Return a valid element ID.

        This does not check for holes in the element numbering.

        Returns
        -------
        int
            An available element ID
        """
        return len(self.elements)

    def get_node_by_id(self, id_: int) -> MeshNode:
        """Return the node with the given ID.

        Paramters
        ---------
        id_: int
            The node ID to return

        Raises
        ------
        ValueError
            Raised if the given node ID does not exist
        """
        for node in self._nodes:
            if node.id_ == id_:
                return node
        raise ValueError(f'No node with ID {id_} found')

    def get_node(self, point: Union[Point2D, Point3D], *,
                 search_effort: float = 1.0) -> MeshNode:
        """Return the node closest to the given point.

        By default, this exhaustively searches the entire mesh. In this
        case, the search is guaranteed to return the closest node.

        The search_effort argument may be used to reduce search effort
        by quitting after X ratio of all nodes have been checked, then
        picking the closest one found.
        This is only useful with space-aware containers where only
        searching ~10% of the entire data set might give sufficiently
        high confidence.

        If a 3D point is passed, only the first two coordinates will be
        used for containment checking.

        Parameters
        ----------
        point : Union[Tuple[float, float], Tuple[float, float, float]]
            The point to check
        search_effort : float, optional
            The ratio of the data to check, by default 1.0

        Returns
        -------
        MeshNode
            The closest node to the given point

        Raises
        ------
        ValueError
            Raised if the search_effort argument valls is equal to or
            less than 0, or if it is greater than 1
        ValueError
            Raised if the mesh is empty
        """
        if search_effort <= 0.0 or search_effort > 1.0:
            raise ValueError('Search effort must fall within the (0.0, 1.0] '
                             f'range (current: {search_effort}')
        if not self.nodes:
            raise ValueError('Mesh does not contain any nodes')

        point_2d: Point2D = point[:2]
        lowest_dist = -1.0
        closest_node: Optional[MeshNode] = None
        for index, node in enumerate(self.nodes.iter_spatial(point_2d)):
            dist = dist_2d(node.spatial_marker, point_2d)
            if lowest_dist < 0.0 or dist < lowest_dist:
                lowest_dist = dist
                closest_node = node
            if (index+1) / len(self.nodes) > search_effort:
                logger.info('Exceeded search effort of %f: %d of %d nodes '
                            'checked', search_effort, index+1, len(self.nodes))
                break

        assert closest_node is not None
        return closest_node

    def _get_node_id(self) -> int:
        """Return a valid node ID.

        This does not check for holes in the node numbering.

        Returns
        -------
        int
            An available node ID
        """
        return len(self.nodes)

    def node_at(self, point: Union[Point2D, Point3D]) -> MeshNode:
        """Return the node at the given point.

        If no node is found at the given point, a ValueError is raised.

        This performs an exact check, be wary of floating-point errors.

        To get the closest node to a given point without an exact
        check, use the Mesh.get_node() method instead.

        If a 3D point is passed, only the first two coordinates will be
        used for containment checking.

        Parameters
        ----------
        point : Union[Tuple[float, float], Tuple[float, float, float]]
            The point to check

        Returns
        -------
        MeshNode
            The mesh node at the given point

        Raises
        ------
        ValueError
            Raised if no node was found at the given point
        """
        for node in self.nodes.iter_spatial(point[:2]):
            if node.spatial_marker == point[:2]:
                return node
        raise ValueError(f'No node found at {point}')

    def purge_nodes(self) -> List[MeshNode]:
        """Delete any dangling nodes from the mesh.

        Dangling nodes are nodes that are not referenced by any mesh
        elements.

        Returns
        -------
        List[MeshNode]
            The list of nodes that were removed from the mesh
        """
        purged = [n for n in self.nodes if not n.elements]
        _ = (self.remove_node(n) for n in purged)  # type: ignore
        return purged

    def remove_element(self, element: MeshElement) -> None:
        """Remove an element from the mesh.

        This does not delete the element's defining nodes, only the
        element itself is removed and detached from its defining nodes.

        Parameters
        ----------
        element : MeshElement
            The element to remove

        Raises
        ------
        ValueError
            Raised if the element is not part of this mesh
        """
        try:
            self._elements.remove(element)
        except KeyError as err:
            raise ValueError(
                f'MeshElement {element} is not part of the mesh') from err
        else:
            for node in list(element.nodes):
                node.detach_element(element)

    def remove_node(self, node: MeshNode) -> None:
        """Remove a node from the mesh.

        This also deletes any elements that were defined using the
        deleted node.

        Parameters
        ----------
        node : MeshNode
            The node to delete

        Raises
        ------
        ValueError
            Raised if the node is not part of this mesh
        """
        logger.debug('Deleted node with ID %d, also deleted %d associated '
                     'elements.', node.id_, len(node.elements))
        for element in list(node.elements):
            self.remove_element(element)
        try:
            self._nodes.remove(node)
        except KeyError as err:
            raise ValueError(
                f'MeshNode {node} is not part of the mesh') from err

    def save(self, filename: str, num_materials: Optional[int] = None) -> None:
        """Save the current mesh as a 2DM mesh file.

        Parameters
        ----------
        filename : str
            The file path to save the mesh to
        num_materials : int, optional
            The number of materials to write to disk, by default None
        """
        # This dict is used to keep track of any shifts in the node IDs
        transform_ids: Dict[int, int] = {}

        # Read the num_materials field from the first element if not provided
        num_materials = len(next(iter(self.elements)).materials)

        with py2dm.Writer(filename) as mesh:
            # Specify the number of materials to write and expect
            mesh.num_materials_per_elem(num_materials)
            mesh.write_header()

            # Write nodes
            nodes: List[MeshNode] = counting_sort(
                list(self.nodes), lambda x: x.id_)
            for node in nodes:
                transform_ids[node.id_] = mesh.node(*node.as_tuple())
            mesh.write_nodes()

            # Write elements
            elements: List[MeshElement] = counting_sort(
                list(self.elements), lambda x: x.id_)
            for element in elements:
                node_ids = tuple(transform_ids[n.id_] for n in element.nodes)
                mesh.element(py2dm.Element3T, node_ids,
                             element.materials[:num_materials])
            mesh.write_elements()

            # Write node strings
            for name, nodes in self.node_strings.items():
                node_ids = tuple(transform_ids[n.id_] for n in nodes)
                mesh.node_string(node_ids, name)
            mesh.write_node_strings()

        # BASEMENT 3 compatibility: Remove trailing whitespace
        self._strip_trailing_newline(filename)

    @staticmethod
    def _strip_trailing_newline(filename: str) -> None:
        """Remove any trailing newlines in the given file.

        This is necessary for compatibility with BASEMENT 3. The file
        is truncated in-place.

        Parameters
        ----------
        filename : str
            The file to truncate
        """
        with open(filename, 'rb+') as file_:
            # Jump to the end of the file
            file_.seek(0, os.SEEK_END)
            # This loop reads char by char from the back of the file as long as the
            # characters returned are considered whitespace.
            while not file_.read(1).strip():
                file_.seek(-2, os.SEEK_CUR)
            # The last read bit was non-whitespace; truncate the file
            file_.truncate()



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Objects related to Triangle geometries.

This mainly covers the 2D equivalent of mesh objects like nodes,
line segments and mesh elements.

All objects defined herein are prefixed with "PSLG" to differentiate
them from their 3D equivalents used throughout the rest of BASEmesh.
"""

from collections import OrderedDict
from typing import Any, Dict, List, Optional, Tuple


class PSLGNode:
    """A two-dimensional PSLG node for use in Triangle."""

    __slots__ = ('_attributes', '_id', '_pos_x', '_pos_y')

    def __init__(self, pos_x: float, pos_y: float, **kwargs: Any) -> None:
        """Instantiate a new PSLG node.

        Note that the default initializer is mostly intended for
        specifying the input parameters for triangulation.
        Nodes read from a Triangle .node-file will be parsed by the
        from_string factory method, which will also populate the ID.
        The ID of manually initialized nodes will be None.

        Extra keyword arguments will be stored along with the node and
        can be used to preserve node information during triangulation,
        such as the Z-coordinates of 3D points.

        Parameters
        ----------
        pos_x : float
            X position of the node
        pos_y : float
            Y position of the node
        """
        self._id: Optional[int] = kwargs.pop('_init_set_id', None)
        self._attributes = OrderedDict(
            sorted(kwargs.items(), key=lambda x: x[0].lower()))
        self._pos_x = pos_x
        self._pos_y = pos_y

    @classmethod
    def from_string(cls, string: str, *, use_boundary: bool = False,
                    column_names: Optional[List[str]] = None) -> 'PSLGNode':
        """Parse a Triangle output line and return a node.

        The optional column_names kwarg can be used to restore the
        attribute names stored along with the node. If this list is not
        provided, attribute names will be generated as "Attr_1",
        "Attr_2", etc.

        Parameters
        ----------
        string : str
            The string to parse
        use_boundary : bool, optional
            Whether boundary markers were included in the line; set
            this to True, by default False
        column_names : List[str], optional
            Names used to resolve the positional node attributes in the
            string to named attributes, by default None

        Returns
        -------
        PSLGNode
            The node represented by the string

        Raises
        ------
        ValueError
            Raised if the number of attributes in column_names is less
            than the number of attribute values found
        """
        # NOTE: Node table format:
        # <Node ID> <X> <Y> [<Attr 1> <Attr 2> ...] [<boundary>]
        words = string.split()
        id_ = int(words[0])
        pos = tuple(float(words[i]) for i in range(1, 3))
        attr_values = words[3:-1] if use_boundary else words[3:]
        # Find/set attribute names
        if column_names is None:
            column_names = [f'Attr_{i+1}' for i, _ in enumerate(attr_values)]
        else:
            column_names = sorted(column_names)
            if len(column_names) != len(attr_values):
                raise ValueError(f'{len(column_names)} column names provided '
                                 f'but {len(attr_values)} values encountered')
        attr_dict = {a: attr_values[i] for i, a in enumerate(column_names)}
        # Instantiate and return the node
        instance = cls(pos[0], pos[1], _init_set_id=id_, **attr_dict)
        return instance

    @property
    def attributes(self) -> Dict[str, str]:
        """Return a dict of values that were stored with the node."""
        return self._attributes

    @property
    def id_(self) -> Optional[int]:
        """Return the unique ID for this node.

        Note that this will only return a valid integer for nodes that
        were read from Triangle output files. Manually defined nodes
        will instead return None.
        """
        return self._id

    def as_tuple(self) -> Tuple[float, float]:
        """Return the coordinates of the node as a tuple of floats."""
        return self._pos_x, self._pos_y


class PSLGSegment:
    """A two-dimensional PSLG segment for use in Triangle."""

    __slots__ = ('_end', '_id', '_start')

    def __init__(self, start: PSLGNode, end: PSLGNode, **kwargs: Any) -> None:
        """Initialise a new PSLG segment.

        Note that the default initializer is mostly intended for
        specifying the input parameters for triangulation.
        Segments read from a Triangle .poly-file will be parsed by
        the from_string factory method, which will also populate the
        ID. The ID of manually initialized segments will be None.

        Parameters
        ----------
        start : PSLGNode
            The starting node of the segment
        end : PSLGNode
            The end node of the segment

        Raises
        ------
        ValueError
            Raised if any unexpected keyword arguments are passed
        """
        self._end = end
        self._id: Optional[int] = kwargs.pop('_init_set_id', None)
        self._start = start
        # NOTE: No keyword argeuments beyond the internal initializer are
        # supported
        if kwargs:
            raise ValueError(
                f'Unexpected keyword argument {next(iter(kwargs))}')

    @classmethod
    def from_points(cls, start_point: Tuple[float, ...],
                    end_point: Tuple[float, ...], *,
                    nodes: Optional[List[PSLGNode]] = None) -> 'PSLGSegment':
        """Create a new segment from the given points.

         Note that only the first two dimensions of the input points
        will be honoured, any additional coordinates are ignored.

        As with the default constructor, the ID will always be None.

        Parameters
        ----------
        start_point : Tuple[float, ...]
            The coordinates of the beginning of the line
        end_point : Tuple[float, ...]
            The coordinates of the end of the line
        nodes : List[PSLGNode], optional
            An optional list of nodes to check the start and end point
            against; if they match, they will be used instead of a new
            node being created, by default None

        Returns
        -------
        PSLGSegment
            The segment that was created
        """
        start = None
        end = None
        if nodes is not None:
            # Search the nodes input list for the given input points
            for node in nodes:
                if start_point[:2] == node.as_tuple():
                    start = node
                if end_point[:2] == node.as_tuple():
                    end = node
                # If both points have been matched to nodes, exit the loop
                if start is not None and end is not None:
                    break
        # Create new nodes for any points that could not be matched to existing
        # nodes
        if start is None:
            # Add Z attribute if possible
            if len(start_point) > 2:
                start = PSLGNode(*start_point[:2], pos_z=start_point[2])
            else:
                start = PSLGNode(*start_point[:2])
            if nodes is not None:
                nodes.append(start)
        if end is None:
            # Add Z attribute if possible
            if len(end_point) > 2:
                end = PSLGNode(*end_point[:2], pos_z=end_point[2])
            else:
                end = PSLGNode(*end_point[:2])
            if nodes is not None:
                nodes.append(end)
        return PSLGSegment(start, end)

    @classmethod
    def from_string(cls, string: str, nodes: List[PSLGNode]) -> 'PSLGSegment':
        """Parse a Triangle output line and return a segment.

        Parameters
        ----------
        string : str
            The string to parse
        nodes : List[PSLGNode]
            The sorted list of nodes to link the segment to

        Returns
        -------
        PSLGSegment
            The segment represented by the string
        """
        # NOTE: Segment table format:
        # <Segment ID> <Node #1 ID> <Node #2 ID> [<boundary>]
        words = string.split()
        id_ = int(words[0])
        node_ids = tuple(int(i) for i in range(1, 3))
        segment_nodes = [nodes[i] for i in node_ids]
        # Instantiate and return the segment
        return cls(segment_nodes[0], segment_nodes[1], _init_set_id=id_)

    @property
    def end(self) -> PSLGNode:
        """Return the node defining the end of the segment."""
        return self._end

    @property
    def id_(self) -> Optional[int]:
        """Return the unique ID of this segment.

        Note that this will only return a valid integer for segments
        that were read from Triangle output files. Manually defined
        segments will instead return None.
        """
        return self._id

    @property
    def as_tuple(self) -> Tuple[Tuple[float, float], Tuple[float, float]]:
        """Return the start and end point as two tuples of floats."""
        return self.start.as_tuple()[:2], self.end.as_tuple()[:2]

    @property
    def start(self) -> PSLGNode:
        """Return the node defining the start of the segment."""
        return self._start


class PSLGElement:
    """A two-dimensional PSLG element for use in Triangle."""

    __slots__ = ('_id', '_matid', '_node1', '_node2', '_node3')

    def __init__(self, node1: PSLGNode, node2: PSLGNode, node3: PSLGNode,
                 matid: Optional[int] = None, **kwargs: Any) -> None:
        """Initialise a new PSLG element.

        Note that the default initializer is mostly intended for
        specifying the input parameters for triangulation.
        Elements read from a Triangle .ele-file will be parsed by the
        from_string factory method, which will also populate the ID.
        The ID of manually initialized elements will be None.

        Parameters
        ----------
        node1 : PSLGNode
            The first node of the element
        node2 : PSLGNode
            The second node of the element
        node3 : PSLGNode
            The third node of the element
        matid : int, optional
            The material ID for this element, by default None
        """
        self._id: Optional[int] = kwargs.pop('_init_set_id', None)
        self._matid = matid
        self._node1 = node1
        self._node2 = node2
        self._node3 = node3

    @classmethod
    def from_string(cls, string: str, nodes: List[PSLGNode], *,
                    use_matid: bool = True) -> 'PSLGElement':
        """Parse a Triangle output line and return an element.

        If use_matid is True, this will expect a MATID column in the
        line.

        Returns
        -------
        PSLGElement
            The element represented by the string

        Raises
        ------
        ValueError
            Raised if use_matid is True but no MATID could be found
        """
        # NOTE: Element table format:
        # <Element ID> <N#1 ID> <N#2 ID> <N#3 ID> [<Attr>]
        words = string.split()
        id_ = int(words[0])
        node_ids = tuple(int(words[n]) for n in range(1, 4))
        element_nodes = [nodes[i] for i in node_ids]
        if use_matid:
            try:
                attr_dict = {'matid': int(words[4])}
            except IndexError as err:
                raise ValueError('Missing element attribute') from err
        else:
            attr_dict = {}
        # Instantiate and return the element
        instance = cls(element_nodes[0], element_nodes[1], element_nodes[2],
                       _init_set_id=id_, **attr_dict)
        return instance

    @property
    def id_(self) -> Optional[int]:
        """Return the unique ID for this element.

        Note that this will only return a valid integer for elements
        that were read from Triangle output files. Manually defined
        elements will instead return None.
        """
        return self._id

    @property
    def matid(self) -> Optional[int]:
        """Return the material ID of the element."""
        return self._matid

    @property
    def nodes(self) -> Tuple[PSLGNode, PSLGNode, PSLGNode]:
        """Return the nodes defining the element."""
        return self._node1, self._node2, self._node3



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Channel geometry components and supporting code.

This module defines the Vertex, CrossSection, and ChannelGeometry
classes used by the channel generation utility.
"""

import collections
import copy
import warnings
from typing import Any, Dict, Iterable, List, Optional, Tuple

from ..algorithms import rotate_2d
from ..core import Mesh
from ..interpolation import interpolate_mesh
from ..meshing import elevation_mesh, quality_mesh
from ..triangle import PSLGNode, PSLGSegment
from ..types import Line3D, Point2D, Point3D
from .io import BaseChainReader, BaseChainCrossSection, BaseChainSoilDef


class Vertex:
    """A single vertex used to define a channel cross section.

    Vertices live in the cross section's local coordinate system, with
    the X axis pointing right and the Y axis pointing up for an
    observer facing downstream. The origin lies in the leftmost vertex
    for a given cross section.

    In addition to its relative location, a vertex also may have a
    constraint associated with it. This can be any object supporting
    equality checks and will be used to dynamically create break lines
    between two cross sections.

    See the :meth:`CrossSection.process_constraints()` method for
    details on the constraint system.

    Attributes
    ----------
    pos : Tuple[Tuple[float, float], Tuple[float, float]]
        The cross-section-relative position of the vertex.
    constraint : Any
        An object used to generate break lines to neighbouring cross
        sections.

    """

    def __init__(self, pos_x: float, pos_y: float,
                 constraint: Any = None) -> None:
        self.pos: Point2D = (pos_x, pos_y)
        self.constraint: Any = constraint

    def __repr__(self) -> str:
        return f'<Vertex at {self.pos}>'

    def to_world_coords(self, anchor: Point3D, angle: float = 0.0) -> Point3D:
        """Convert the vertex position to world space coordinates.

        This performs the translation between the local cross section
        coordinate system and the global coordinate system.

        Parameters
        ----------
        anchor : Point3D
            The leftmost point in the cross section. This point acts as
            the origin for the cross section's local CRS. Any height
            offsets for the cross sections should be included in this
            point's Z offset.
        angle : float, optional
            The global angle of the given cross section, by default
            0.0.

        Returns
        -------
        Tuple[float, float, float]
            The global world space position of the vertex.
        """
        pos = anchor[0] + self.pos[0], anchor[1]
        # Rotate the vertex around the anchor, if needed
        if angle != 0.0:
            pos = rotate_2d(pos, angle, (anchor[0], anchor[1]))
        # Calculate the elevation of the vertex
        elevation = anchor[2] + self.pos[1]
        return tuple(round(f, 12) for f in (*pos, elevation))  # type: ignore


class CrossSection:
    """A cross section making up a channel geometry.

    A cross section's vertices are defined relative to its origin. When
    looking at a cross section from the front (i.e. in river flow
    direction), the first coordinate (X) will point to the right, with
    the second (Y) pointing up.
    This means that the Y coordinate in the cross section's coordinate
    system will map to the Z coordinate (elevation) in the global
    coordinate system.

    Depending on the data provided to the :class:`CrossSection`
    instance, there exist multiple ways to convert its local coordinate
    system into the global CRS.

    :meth:`position_relative()` will position the cross section
    relative to a previous cross section using the difference in the
    :attr:`distance_coord` attribute for distance, as well as the
    respective cross sections' :attr:`angle` attributes.

    For cross sections with absolute positioning information available
    through their :attr:`anchor` attribute, you can alternatively use
    :meth:`position_absolute()`, which will use the global position
    to position the cross section while ignoring :attr:`distance_coord`
    altogether. This is analogous to the ``left_point_global_coord``
    field in the BASEchain geometry (BMG) format.

    Attributes
    ----------
    name : str
        A unique name used to identify this cross section.
    vertices : List[Vertex]
        The list of vertices defined for this cross section.
    flow_axis_coord : float
        The absolute downstream positioning of the cross section along
        the river flow axis

        .. seealso::
            :attr:`distance_coord`
    flow_axis_shift : float, optional
        Relative position of the river flow axis within the cross
        section as a real value between 0.0 and 1.0. This is used to
        generate the river flow axis polyline, as well as to improve
        the relative positioning of cross sections, by default 0.5
    anchor : Optional[Point2D], optional
        The absolute position of the first (i.e. leftmost) vertex in
        the cross section, used as the origin of the cross section's
        local coordinate system. This value is used for absolute
        positioning of the cross section.

        .. seealso::
            :attr:`angle`
    reference_height : float, optional
        An elevation offset applied to all node elevations, by default
        0.0
    angle : float, optional
        The angle of the cross section as measured counter-clockwise
        with the cross section anchor point as the pivot. This angle is
        relative to the global coordinate system, with an angle of 0°
        corresponding to a cross section facing the global ``(0, 1)``
        vector, by default 0.0
    sternberg : float, optional
        Sternberg factor for the reduction of volume by abrasion, by
        default 0.0
    bed_form_factor : float, optional
        Factor considering the influence of bed forms on bottom shear
        stress (ripple factor). This factor reduces the bottom shear
        stress due to bedforms with a constant value. If this factor is
        set to 1.0 (default), there is no influence and the standard
        shear stress is used, by default 1.0
        .. seealso:: :attr:`theta_critic`
    theta_critic : float, optional
        User-defined dimensionless critical bottom shear stress
        (Shields parameter) for this cross section. If specified, this
        local theta critic is used for this cross section. Permissible
        values range from 0.0 to 1.0. Set to a negative value to use
        the standard theta critical as defined in the
        ``BEDLOAD_PARAMETER`` block, by default -1.0
        .. seealso:: :attr:`bed_form_factor`
    friction_calibration_factor : float, optional
        Calibration factor between 0.0 and 10.0 to locally modify the
        friction value for this cross section, by default 1.0
    default_friction : Optional[float], optional
        Default Strickler friction value to use for this cross section,
        by default None
    friction_slices : Optional[List[Tuple[int, int]]], optional
        The edge indices to which to apply the coefficients defined
        in :attr:`friction_coefficients`, by default None
    friction_coefficients : Optional[List[float]], optional
        Strickler values to apply to the segments in the groups defined
        in :attr:`friction_ranges` or :attr:`friction_slice_indexes`
        respectively, by default None
    interpolation_fixpoints : List[int], optional
        The interpolation fixpoints are used to specify break lines
        in-between existing cross sections. This can be used to insert
        new cross sections, as well as to improve the quality of a 2D
        mesh export.

        The first and last node in a cross section are always
        connected. Nodes are numbered starting at 1, by default None
    table_fixed_points : List[float], optional
        Fixed points which should be included in the table calculation.
        Note that these values are only used if you use lookup tables
        for hydraulic properties (see ``SECTION_COMPUTATION`` in the
        BASEMENT command file documentation for details), by default
        None
    table_fixed_lower_limit : float, optional
        Allows to fix the lowest water surface elevation used in the
        lookup table for ``A(z)`` to a different value than the lowest
        elevation in the cross section. This option is useful if you
        include a deep lake in your model containing a large water
        volume not participating in the dynamics of your river system,
        by default None
    active_slice : Optional[Tuple[int, int]], optional
        This range defines the active part of the cross section and
        must span from the left to the right dike. If not specified,
        the whole cross section is assumed to be active, and the
        elevation of the first and last point will be assumed to
        correspond to the dike elevation, by default None
    bottom_slice : Optional[Tuple[int, int]], optional
        Range defining the span of the bottom/bedload active zone, by
        default None
    main_channel_slice : Optional[Tuple[int, int]], optional
        Range defining the span of the main channel. Everything outside
        of the main channel is treated as floodplain. If not defined,
        the whole cross section is assumed to belong to the main
        channel, by default None
    water_flow_slice : Optional[Tuple[int, int]], optional
        Used to limit the part of the part of the cross section that
        will be used for solving of the momentum equation, the rest of
        the cross section is considered to be only storage area in the
        continuity equation. By default, the whole cross section area
        is flowing.

        Avoid defining storage areas in the main channel as this can
        cause instability if only the non-flowing part of the channel
        is wetted.

        Additionally, the non-flowing areas are not taken into account
        for sediment deposition and erosion, by default None
    soil_defs : Dict[int, Tuple[int, int]], optional
        A dictionary mapping soil definition indices to their edge's
        slice index.

    """

    def __init__(self, name: str,
                 vertices: Optional[Iterable[Vertex]] = None) -> None:
        self.name = name
        self.vertices: List[Vertex] = (
            [] if vertices is None else list(vertices))
        self.flow_axis_coord: float = 0.0
        self.flow_axis_shift: float = 0.5

        self.anchor: Optional[Point2D] = None
        self.reference_height: float = 0.0
        self.angle: float = 0.0

        self.sternberg: float = 0.0
        self.bed_form_factor: float = 1.0
        self.theta_critic: float = -1.0

        self.friction_calibration_factor: float = 1.0
        self.default_friction: Optional[float] = None
        self.friction_slices: Optional[List[Tuple[int, int]]] = None
        self.friction_coefficients: Optional[List[float]] = None

        self.interpolation_fixpoints: Optional[List[int]] = None
        self.table_fixed_points: Optional[List[float]] = None
        self.table_fixed_lower_limit: Optional[float] = None

        self.active_slice: Optional[Tuple[int, int]] = None
        self.bottom_slice: Optional[Tuple[int, int]] = None
        self.main_channel_slice: Optional[Tuple[int, int]] = None
        self.water_flow_slice: Optional[Tuple[int, int]] = None

        self.soil_defs: Dict[int, Tuple[int, int]] = {}

    @property
    def distance_coord(self) -> float:
        """Distance coord of the cross section.

        Measured up- to downstream, the unit is kilometres.

        .. seealso::
            :attr:`flow_axis_coord`

        """
        return self.flow_axis_coord / 1000.0

    @distance_coord.setter
    def distance_coord(self, value: float) -> None:
        """Setter for :attr:`distance_coord`."""
        self.flow_axis_coord = value * 1000.0

    @property
    def flow_axis(self) -> Point2D:
        """Return the 2D point defining the cross section's flow axis.

        The returned point is given in the local coordinate system.
        """
        origin = self.vertices[0].pos
        return origin[0] + self.length*self.flow_axis_shift, 0.0

    @property
    def geo_referenced(self) -> bool:
        """Whether the cross section is georeferenced.

        If True, the cross section can be converted back into a list of
        3D coordinates using the :meth:`position_absolute()` method.
        """
        return self.anchor is not None

    @property
    def length(self) -> float:
        """Return the length of the cross section."""
        return self.vertices[-1].pos[0] - self.vertices[0].pos[0]

    @property
    def span(self) -> float:
        """Like :attr:`length` but includes first point offset.

        If the first point in the cross section has a non-zero X
        coordinate, this will be included in the cross section's span,
        but not its length.
        """
        return self.vertices[-1].pos[0]

    def clone(self, deep_copy: bool = False) -> 'CrossSection':
        """Return a duplicate of the current cross section.

        Since cross sections also define their relative location in the
        global world space, re-using them is prone to cause issues as
        they both refer to the same reference.

        You can use this method to create a clone of a cross section,
        which will copy the set of vertices and distance constraints.

        By default, this will keep the same vertex objects and only
        update the mutable list containing them. For a proper deep
        copy, set the "deep_copy" argument to True.

        Parameters
        ----------
        deep_copy : bool, optional
            By default, internal objects like the Vertex instances or
            their constraints are not copied, only referenced in a new
            list. Set this flag to True to perform a deep copy instead.

        Returns
        -------
        CrossSection
            A copy of the current cross section.

        """
        if deep_copy:
            return copy.deepcopy(self)
        return copy.copy(self)

    def position_absolute(self) -> List[Point3D]:
        """Position the cross section in 3D space.

        This method positions cross sections using their absolute
        position information.

        .. seealso::
            :meth:`position_relative`

        Raises
        ------
        RuntimeError
            Raised if no absolute positioning information has been
            provided for this cross section.

        Returns
        -------
        List[Tuple[float, float, float]]
            The converted vertices in the global coordinate system.

        """
        if self.anchor is None:
            raise RuntimeError(
                'Unable to generate absolute position for Cross section '
                f'{self.name}: no anchor defined; this likely means that '
                'left_point_global_coords is not defined for the associated '
                'BMG')
        anchor_3d = *self.anchor, self.reference_height
        return [v.to_world_coords(anchor_3d, self.angle)
                for v in self.vertices]

    def position_relative(self) -> List[Point3D]:
        """Position the cross section in 3D space.

        This method positions cross sections relative to a previous
        cross section.

        .. seealso::
            :meth:`position_absolute`

        Returns
        -------
        List[Tuple[float, float, float]]
            The converted vertices in the global coordinate system.

        """
        # Get the flow axis position for the current section
        flow_axis = self.flow_axis_coord, 0.0
        # Get the current section's anchor point
        offset = self.span - self.length + self.length*self.flow_axis_shift
        anchor = flow_axis[0], flow_axis[1] - offset
        # Position the current section's vertices based on this anchor
        return [v.to_world_coords((*anchor, self.reference_height), 90.0)
                for v in self.vertices]

    @classmethod
    def from_basechain(cls, data: BaseChainCrossSection) -> 'CrossSection':
        """Create a new cross section from a BMG data class."""
        cs_ = cls(data.name, (Vertex(*pos) for pos in data.node_coords))
        cs_.distance_coord = data.distance_coord
        cs_.sternberg = data.sternberg
        cs_.reference_height = data.reference_height
        if data.left_point_global_coords is not None:
            *anchor, cs_.reference_height = data.left_point_global_coords
            cs_.anchor = tuple(anchor)  # type: ignore
            # left_point_global_coords include the relative elevation of the
            # leftmost vertex which must be removed
            cs_.reference_height -= cs_.vertices[0].pos[1]
            # Check for conflicting data
            difference = cs_.reference_height - data.reference_height
            if difference > 1e-10:
                warnings.warn(
                    'Reference height mismatch for cross section '
                    f'"{cs_.name}": {cs_.reference_height}'
                    f'!= {data.reference_height}\n'
                    f'Using global value {cs_.reference_height}')
        cs_.angle = data.orientation_angle
        cs_.interpolation_fixpoints = data.interpolation_fixpoints
        # Water flow range
        if data.water_flow_slice_indexes is not None:
            cs_.water_flow_slice = data.water_flow_slice_indexes
            if data.water_flow_range is not None:
                warnings.warn('water_flow_range value is redundant, ignoring')
        elif data.water_flow_range is not None:
            cs_.water_flow_slice = cs_.range_to_slice(
                data.water_flow_range)
        # Main channel range
        if data.main_channel_slice_indexes is not None:
            cs_.main_channel_slice = data.main_channel_slice_indexes
            if data.main_channel_range is not None:
                warnings.warn(
                    'main_channel_range value is redundant, ignoring')
        elif data.main_channel_range is not None:
            cs_.main_channel_slice = cs_.range_to_slice(
                data.main_channel_range)
        # Bottom range
        if data.bottom_slice_indexes is not None:
            cs_.bottom_slice = data.bottom_slice_indexes
            if data.bottom_range is not None:
                warnings.warn('bottom_range value is redundant, ignoring')
        elif data.bottom_range is not None:
            cs_.bottom_slice = cs_.range_to_slice(data.bottom_range)
        # Active range
        if data.active_slice_indexes is not None:
            cs_.active_slice = data.active_slice_indexes
            if data.active_range is not None:
                warnings.warn('active_range value is redundant, ignoring')
        elif data.active_range is not None:
            cs_.active_slice = cs_.range_to_slice(data.active_range)
        cs_.bed_form_factor = data.bed_form_factor
        cs_.theta_critic = data.theta_critic
        cs_.friction_calibration_factor = data.friction_calibration_factor
        cs_.default_friction = data.default_friction
        cs_.friction_coefficients = data.friction_coefficients
        # Friction ranges
        if data.friction_slice_indexes is not None:
            cs_.friction_slices = data.friction_slice_indexes
            if data.friction_ranges is not None:
                warnings.warn('friction_ranges value is redundant, ignoring')
        elif data.friction_ranges is not None:
            cs_.friction_slices = [cs_.range_to_slice(d)
                                   for d in data.friction_ranges]
        cs_.table_fixed_points = data.table_fixed_points
        cs_.table_fixed_lower_limit = data.table_fixed_lower_limit
        # Soil definitions
        if data.soil_defs is not None:
            for def_ in data.soil_defs:
                index = def_.index
                if def_.slice_indexes is not None:
                    nodes = def_.slice_indexes
                elif def_.range is not None:
                    nodes = cs_.range_to_slice(def_.range)
                else:
                    warnings.warn(
                        f'Undefined soil definition ignored: {index}')
                cs_.soil_defs[index] = nodes
        return cs_

    def process_constraints(self, other: 'CrossSection'
                            ) -> List[Tuple[int, int]]:
        """Generate the break lines between two cross sections.

        This will find any vertices that have matching constraints
        across the two cross sections. These vertices are then returned
        as pairs of indices representing the vertex in the first and
        second cross section.

        Parameters
        ----------
        other : CrossSection
            The next cross section in the channel. The break lines will
            be generated between that cross section and the current
            instance (i.e. self).

        Returns
        -------
        List[Tuple[int, int]]
            A list of pairs of indices representing the vertices to
            connect via break lines.
        """
        connections: List[Tuple[int, int]] = []
        for idx_a, vtx_a in enumerate(self.vertices):
            try:
                constraints_a = set(vtx_a.constraint)
            except TypeError:
                constraints_a = {vtx_a.constraint}
            for idx_b, vtx_b in enumerate(other.vertices):
                try:
                    constraints_b = set(vtx_b.constraint)
                except TypeError:
                    constraints_b = {vtx_b.constraint}
                if vtx_a.constraint != vtx_b.constraint:
                    if not set(constraints_a).intersection(set(constraints_b)):
                        continue
                if vtx_a.constraint is None or vtx_b.constraint is None:
                    continue
                connections.append((idx_a, idx_b))
        return connections

    def range_to_slice(self, range_: Tuple[float, float]) -> Tuple[int, int]:
        """Convert a range into explicit slice indices."""
        x_coords = [v.pos[0] for v in self.vertices]
        return x_coords.index(range_[0]), x_coords.index(range_[1])

    def to_basechain(self) -> BaseChainCrossSection:
        """Convert the cross section into a BMG dataclass.

        Returns
        -------
        BaseChainCrossSection
            The BMG cross section equivalent of this section.

        """
        cs_ = BaseChainCrossSection(
            name=self.name,
            distance_coord=self.distance_coord,
            node_coords=[v.pos for v in self.vertices],
            sternberg=self.sternberg,
            reference_height=self.reference_height,
            orientation_angle=self.angle,
            interpolation_fixpoints=self.interpolation_fixpoints,
            water_flow_slice_indexes=self.water_flow_slice,
            main_channel_slice_indexes=self.main_channel_slice,
            bottom_slice_indexes=self.bottom_slice,
            active_slice_indexes=self.active_slice,
            bed_form_factor=self.bed_form_factor,
            theta_critic=self.theta_critic,
            friction_calibration_factor=self.friction_calibration_factor,
            default_friction=self.default_friction,
            friction_coefficients=self.friction_coefficients,
            friction_slice_indexes=self.friction_slices,
            table_fixed_points=self.table_fixed_points,
            table_fixed_lower_limit=self.table_fixed_lower_limit)
        if self.anchor is not None:
            cs_.left_point_global_coords = (
                *self.anchor, self.reference_height+self.vertices[0].pos[1])
        if self.soil_defs:
            cs_.soil_defs = []
            for index, sd_indices in self.soil_defs.items():
                cs_.soil_defs.append(
                    BaseChainSoilDef(index, slice_indexes=sd_indices))
        return cs_


class ChannelGeometry:
    """A BASEchange channel geometry.

    Channel geometries consist of any number of flat, two-dimensional
    cross sections. These cross sections use a local coordinate system
    for their nodes, which is translated into world coordinate system
    as needed.

    Attributes
    ----------
    name : str
        The name of the model. Used to generate custom file names;
        helpful when working with multiple geometries.
    cross_sections : List[CrossSection]
        The list of cross sections defining the channel geometry.
    """

    def __init__(self, name: str = 'Unnamed model',
                 cross_sections: Optional[Iterable[CrossSection]] = None
                 ) -> None:
        self.name = name
        self.cross_sections: List[CrossSection] = (
            [] if cross_sections is None else list(cross_sections))

    def add_cross_section(self, cross_section: CrossSection,
                          insert_at: int = -1) -> None:
        """Add an existing cross section to the channel.

        The optional index argument may be used to specify the index to
        insert the channel at. An insertion index of 2 means that the
        new index 2 will be the inserted element, with all following
        sections (including the original index 2) being shifted by 1.

        Parameters
        ----------
        section : CrossSection
            An existing cross section to add to the channel.
        insert_at : int, optional
            The index to insert the section at, by default -1.
        """
        if insert_at < 0:
            self.cross_sections.append(cross_section)
        else:
            self.cross_sections.insert(insert_at, cross_section)

    @classmethod
    def from_basechain(cls, filepath: str,
                       flow_axis_shift: float = 0.5) -> 'ChannelGeometry':
        """Create a channel geometry from a BASEchain geometry file."""
        # Create a geometry to parse the file into
        instance = cls()
        with BaseChainReader(filepath) as parser:
            for data in parser.cross_sections():
                cs_ = CrossSection.from_basechain(data)
                cs_.flow_axis_shift = flow_axis_shift
                instance.cross_sections.append(cs_)
        return instance

    def new_cross_section(self, name: str, nodes: Iterable[Vertex],
                          insert_at: int = -1) -> CrossSection:
        """Create and add a new cross section to the channel.

        The optional index argument may be used to specify the index to
        insert the channel at. An insertion index of 2 means that the
        new index 2 will be the inserted element, with all following
        sections (including the original index 2) being shifted by 1.

        Parameters
        ----------
        name : str
            The unique name of the cross section
        nodes : Iterable[Vertex]
            Any number of vertices to create the cross section from.
        insert_at : int, optional
            The index to insert the section at, by default -1.
        """
        cs_ = CrossSection(name, nodes)
        self.cross_sections.insert(insert_at, cs_)
        return cs_

    def remove_cross_section(self, index: int) -> CrossSection:
        """Remove and return a cross section from the channel.

        Parameters
        ----------
        index : int
            The index of the cross section to remove.
        """
        return self.cross_sections.pop(index)

    def to_bmg(self) -> List[BaseChainCrossSection]:
        """Convert the geometry into a list of BMG cross sections."""
        return [s.to_basechain() for s in self.cross_sections]

    def to_mesh(self, absolute_pos: bool = False, **kwargs: Any) -> Mesh:
        """Export the channel geometry as a mesh object.

        This converts the channel geometry into a BASEmesh Mesh
        instance that can then be saved for use in BASEMENT.
        """
        # Convert cross sections to 3D line strings
        nodes: List[List[Point3D]] = []
        discarded = 0

        section_nodes: Dict[Point2D, float] = collections.OrderedDict()
        for cs_ in self.cross_sections:
            if absolute_pos:
                section_pos = cs_.position_absolute()
            else:
                section_pos = cs_.position_relative()

            for *vtx, elev in section_pos:
                pos: Point2D = tuple(vtx)  # type: ignore
                if pos in section_nodes:
                    discarded += 1
                    other_elev = section_nodes[pos]
                    if elev > other_elev:
                        section_nodes[pos] = elev
                else:
                    section_nodes[pos] = elev

            nodes.append([(*k, v) for k, v in section_nodes.items()])
            section_nodes.clear()
        if discarded > 0:
            warnings.warn(f'{discarded} nodes were discarded; note that 2D '
                          'meshes cannot describe vertical walls')
        # Process constraints
        break_lines: List[Line3D] = []
        for index, this_cs in enumerate(self.cross_sections[:-1]):
            # Add cross section break line
            for vtx_i in range(len(nodes[index])-1):
                break_lines.append(
                    (nodes[index][vtx_i], nodes[index][vtx_i+1]))
            # Add side borders
            break_lines.append((nodes[index][0], nodes[index+1][0]))
            break_lines.append((nodes[index][-1], nodes[index+1][-1]))
            # Process constraints
            next_cs = self.cross_sections[index+1]
            constraints = this_cs.process_constraints(next_cs)
            for con_a, con_b in constraints:
                line = nodes[index][con_a], nodes[index+1][con_b]
                break_lines.append(line)
        break_lines.append((nodes[-1][0], nodes[-1][-1]))

        # Generate Triangle input geometries
        all_nodes = [n for c in nodes for n in c]
        triangle_nodes = [PSLGNode(*n[:2], pos_z=n[2]) for n in all_nodes]
        triangle_segments = [
            PSLGSegment.from_points(l1[:2], l2[:2], nodes=triangle_nodes)
            for l1, l2 in break_lines]
        # Generate meshes
        em_ = elevation_mesh(triangle_nodes, segments=triangle_segments)
        qm_ = quality_mesh(
            triangle_nodes, segments=triangle_segments, **kwargs)
        interpolate_mesh(qm_, em_)
        return qm_



# A component of the BASEmesh pre-processing toolkit.
# Copyright (C) 2020  ETH Zürich

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Shared utilities for common QGIS operations."""

import os
import warnings
from typing import Callable, Dict, List, Optional, Set, Tuple, Union
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QComboBox
from qgis.core import (QgsApplication, QgsFeature, QgsField,  # type: ignore
                       QgsMapLayer, QgsPoint, QgsProject, QgsVectorLayer,
                       QgsWkbTypes)
from qgis.gui import QgsMapCanvas  # type: ignore
from ..algorithms import split_line
from ..core import Mesh
from ..feedback import Feedback
from ..stringdefs import resolve_string_defs
from ..triangle import HoleMarker, RegionMarker
from ..types import Line2D, Line3D, LineString2D, Point2D, Point3D


def add_layers(canvas: QgsMapCanvas, combo_box: QComboBox, *args: QComboBox,
               filter_: Optional[Callable[[QgsMapLayer], bool]] = None,
               icon: Optional[QIcon] = None) -> None:
    """Add layers from the canvas to the given combo box.

    Parameters
    ----------
    canvas : QgsMapCanvas
        The QGIS layer canvas to process
    combo_box : QComboBox
        The combo box to populate
    filter_ : Callable[[QgsMapLayer], bool], optional
        The function used to select valid layers, by default None
    icon : QIcon, optional
        The layer icon to use for the combo box, by default None
    """
    # Create dummy filter if no filter was provided
    if filter_ is None:
        def yes_man(layer: QgsMapCanvas) -> bool:
            _ = layer  # Value intentionally discarded
            return True
        filter_ = yes_man
    # Add the combo box items
    for layer in canvas.layers():
        if filter_(layer):
            for c_box in (combo_box, *args):
                if icon is None:
                    c_box.addItem(layer.name())
                else:
                    c_box.addItem(icon, layer.name())
                # This value can be retrieved using QComboBox.currentData,
                # saving us having to manually match the layer by name later.
                c_box.setItemData(c_box.count()-1, layer)


def extract_lines(layer: QgsVectorLayer, use_z: bool = True,
                  divcon_field: QgsField = None) -> Set[Union[Line2D, Line3D]]:
    """Process the given layer and return any lines found.

    Depending on the use_z flag, the returned set will contain 2D or 3D
    lines.

    Parameters
    ----------
    layer : QgsVectorLayer
        The layer to process
    use_z : bool, optional
        Whether to return 3D or 2D lines, by default True
    divcon_field : QgsField, optional
        A dividing constraint field used to pre-segmentise lines before
        returning, by default None

    Returns
    -------
    Set[Union[Tuple[Tuple[float, float], Tuple[float, float]],
              Tuple[Tuple[float, float, float],
                    Tuple[float, float, float]]]]
        The lines found in the given layer

    Raises
    ------
    ValueError
        Raised if use_z is True for a 2D layer
    """
    if use_z and not layer_is_3d(layer):
        raise ValueError(f'Layer {layer.name()} is not a 3D layer')
    line_strings = []

    # Get individual line strings
    feature: QgsFeature
    point: Union[Point2D, Point3D]
    for feature in layer.getFeatures():
        line_string = []
        vertex: QgsPoint
        for vertex in feature.geometry().vertices():
            if use_z:
                point = vertex.x(), vertex.y(), vertex.z()
            else:
                point = vertex.x(), vertex.y()
            line_string.append(point)

        line_string = _process_divcon(line_string, feature, divcon_field)

        line_strings.append(tuple(line_string))

    # Break up line strings into unique line segments
    lines = set()
    for line_str in line_strings:
        for index, vertex in enumerate(line_str[:-1]):
            next_vertex = line_str[index+1]
            # Ignore duplicates regardless of orienation
            if not ((vertex, next_vertex) in lines
                    or (next_vertex, vertex) in lines):
                lines.add((vertex, next_vertex))
    return lines  # type: ignore


def extract_points(layer: QgsVectorLayer, use_z: bool = True
                   ) -> Set[Union[Point2D, Point3D]]:
    """Process the given geometry and return any unique points.

    If use_z is True, the returned points will return elevation
    information.

    Parameters
    ----------
    layer : QgsVectorLayer
        The layer to process
    use_z : bool, optional
        Whether to include elevation information in the returned set,
        by default True

    Returns
    -------
    Set[Union[Tuple[float, float], Tuple[float, float, float]]]
        The set of unique points found in the layer

    Raises
    ------
    ValueError
        Raised if use_z is True for a 2D layer
    """
    if use_z and not layer_is_3d(layer):
        raise ValueError(f'Layer {layer.name()} is not a 3D layer')
    points: Set[Union[Point2D, Point3D]] = set()

    feature: QgsFeature
    point: Union[Point2D, Point3D]
    for feature in layer.getFeatures():
        # NOTE: As of 3.10, QgsGeometry.asPoint() always returns a QgsPointXY
        # instances. Iterating over the feature vertices yields us QgsPoint
        # instead.
        vertex: QgsPoint
        for vertex in feature.geometry().vertices():
            if use_z:
                point = vertex.x(), vertex.y(), vertex.z()
            else:
                point = vertex.x(), vertex.y()
            points.add(point)
    return points


def get_plugin_dir() -> str:
    """Return the base directory of the QGIS plugin."""
    profile_dir = os.path.dirname(QgsApplication.qgisUserDatabaseFilePath())
    return os.path.join(profile_dir, 'python', 'plugins', 'basemesh')


def get_project_dir() -> str:
    """Return the current project directory.

    If the project has not yet been saved, the user documents folder is
    returned instead.
    """
    current_project = QgsProject.instance()
    # Return the project directory if it has been saved
    project_dir = str(current_project.absolutePath())
    if project_dir:
        return project_dir
    # Otherwise, return the user documents folder
    if os.name == 'nt':
        user_docs = os.path.expandvars('%USERPROFILE%')
    elif os.name == 'posix':
        user_docs = os.path.expanduser('~')
    else:
        raise RuntimeError(f'Unsupported OS: {os.name}')
    user_docs = os.path.join(user_docs, 'Documents')
    return user_docs


def get_project_title() -> str:
    """Return the current project title.

    If the current project does not have a title specified, its base
    name (i.e. file name without extension) will be used instead.
    If the file has not been saved, the default title "Project1" is
    returned instead.
    """
    project = QgsProject().instance()
    # Return the project's title, if it exists
    if project.title():
        return str(project.title())
    # If not title has been specified, use the project's base name instead
    if project.baseName():
        return str(project.baseName())
    # Use the default file name instead
    default_name = 'Project1'
    project.setTitle(default_name)
    return default_name


def layer_is_3d(layer: QgsVectorLayer) -> bool:
    """Return whether the given layer contains elevation data.

    Parameters
    ----------
    layer : QgsVectorLayer
        The layer to check

    Returns
    -------
    bool
        Whether the layer contains elevation data
    """
    wkb_type = layer.dataProvider().wkbType()
    return bool(QgsWkbTypes.hasZ(wkb_type))


def _process_divcon(line_string: List[Union[Point2D, Point3D]],
                    feat: QgsFeature, divcon_field: QgsField = None
                    ) -> List[Union[Point2D, Point3D]]:
    """Process the dividing constraints in the given line string.

    Parameters
    ----------
    line_string : List[Union[Tuple[float, float],
                             Tuple[float, float, float]]]
        The line string to segmentise
    feat : QgsFeature
        The QGIS feature to check for field values
    divcon_field : QgsField, optional
        The QGIS field to use for segmentisation, by default None

    Returns
    -------
    List[Union[Tuple[float, float], Tuple[float, float, float]]]
        The segmentised line string
    """
    # If no dividing constraint field has been specified, do nothing
    if divcon_field is None:
        return line_string

    # Ensure the dividing constraint field has a sensible value
    attr_index = feat.fields().indexFromName(divcon_field.name())
    segments = int(feat.attributes()[attr_index])
    if segments < 2:
        # No segment -> no splitting required
        return line_string

    # Split every line in the line string into (divcon) elements
    return split_line(line_string, segments)  # type: ignore


def process_markers(layer: QgsVectorLayer, *, area_field: QgsField = None,
                    hole_field: QgsField = None, matid_field: QgsField = None
                    ) -> Tuple[List[RegionMarker], List[HoleMarker]]:
    """Process the given markers layer and return any markers found.

    Parameters
    ----------
    layer : QgsVectorLayer
        The layer to process
    area_field : QgsField, optional
        The max_area field to use, by default None
    hole_field : QgsField, optional
        The is_hole field to use, by default None
    matid_field : QgsField, optional
        The matid field to use, by default None

    Returns
    -------
    Tuple[List[RegionMarker], List[HoleMarker]]
        A list of region and hole markers found in the layer

    Raises
    ------
    ValueError
        Raised if the input layer is not a point layer
    """
    if layer.geometryType() != QgsWkbTypes.PointGeometry:
        raise ValueError(f'Layer "{layer.name()}" is not a point layer')
    holes = []
    regions = []
    for feat in layer.getFeatures():
        if feat.geometry().type() == QgsWkbTypes.PointGeometry:
            point: QgsPoint = feat.geometry().asPoint()
            pos_x, pos_y = point.x(), point.y()

            # Hole flag
            is_hole = False
            if hole_field is not None:
                attr_index = feat.fields().indexFromName(hole_field.name())
                is_hole = bool(feat.attributes()[attr_index])
                if is_hole:
                    holes.append(HoleMarker(pos_x, pos_y))

            # Maximum area constraint
            max_area = -1.0
            if area_field is not None:
                attr_index = feat.fields().indexFromName(area_field.name())
                max_area = feat.attributes()[attr_index]
            if is_hole and max_area is not None and max_area > 0.0:
                warnings.warn(f'Point {(pos_x, pos_y)} marks a hole, its '
                              f'max area ({max_area}) will have no effect')
            # MATID
            matid = 0
            if matid_field is not None:
                attr_index = feat.fields().indexFromName(matid_field.name())
                matid = feat.attributes()[attr_index]
                if matid is None:
                    matid = 0
            if is_hole and matid is not None:
                warnings.warn(f'Point {(pos_x, pos_y)} marks a hole, its '
                              f'MATID ({matid}) will have no effect')

            if max_area > 0.0 or matid > 0:
                regions.append(RegionMarker(pos_x, pos_y, max_area=max_area,
                                            attribute=matid))
    return regions, holes


def process_string_defs(layer: QgsVectorLayer, id_field: QgsField, mesh: Mesh,
                        precision: float = 0.0, *,
                        feedback: Optional[Feedback] = None
                        ) -> Dict[str, List[int]]:
    """Write the string definitions for the given mesh.

    Parameters
    ----------
    layer : QgsVectorLayer
        The layer containing the string definitions
    id_field : QgsField
        The layer field to use as the string definition name/ID
    mesh : Mesh
        The mesh used to process the line strings
    precision : float, optional
        The maximum distance between a point and its string definition,
        by default 0.0
    feedback : Feedback, optional
        A feedback option used to communicate with the caller, by
        default None

    Returns
    -------
    Dict[str, List[int]]
        A dictionary mapping string definitions to their node IDs
    """
    string_def_lines: Dict[str, LineString2D] = {}
    feature: QgsFeature
    for feature in layer.getFeatures():
        # Get string ID field
        attr_index = feature.fields().indexFromName(id_field.name())
        name = feature.attributes()[attr_index]
        line_string: List[Point2D] = []
        # Add vertices
        vertex: QgsPoint
        for vertex in feature.geometry().vertices():
            line_string.append((vertex.x(), vertex.y()))
        string_def_lines[name] = tuple(line_string)
    return resolve_string_defs(string_def_lines, mesh, precision,
                               feedback=feedback)
